#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from __future__ import annotations

import json
import logging
import os
import re
from typing import Tuple
from zipfile import ZipFile

from filelock import FileLock
from rdflib import URIRef
from tqdm import tqdm

from oc_meta.lib.file_manager import get_csv_data, write_csv
from oc_meta.plugins.editor import MetaEditor

logging.getLogger("filelock").setLevel(logging.INFO)

URI_TYPE_DICT = {
    'http://purl.org/spar/doco/Abstract': 'abstract',
    'http://purl.org/spar/fabio/ArchivalDocument': 'archival document',
    'http://purl.org/spar/fabio/AudioDocument': 'audio document',
    'http://purl.org/spar/fabio/Book': 'book',
    'http://purl.org/spar/fabio/BookChapter': 'book chapter',
    'http://purl.org/spar/fabio/ExpressionCollection': 'book section', 
    'http://purl.org/spar/fabio/BookSeries': 'book series', 
    'http://purl.org/spar/fabio/BookSet': 'book set',
    'http://purl.org/spar/fabio/ComputerProgram': 'computer program',
    'http://purl.org/spar/doco/Part': 'book part',
    'http://purl.org/spar/fabio/Expression': '',
    'http://purl.org/spar/fabio/DataFile': 'dataset',
    'http://purl.org/spar/fabio/DataManagementPlan': 'data management plan',
    'http://purl.org/spar/fabio/Thesis': 'dissertation',
    'http://purl.org/spar/fabio/Editorial': 'editorial',
    'http://purl.org/spar/fabio/Journal': 'journal', 
    'http://purl.org/spar/fabio/JournalArticle': 'journal article',
    'http://purl.org/spar/fabio/JournalEditorial': 'journal editorial',
    'http://purl.org/spar/fabio/JournalIssue': 'journal issue', 
    'http://purl.org/spar/fabio/JournalVolume': 'journal volume',
    'http://purl.org/spar/fabio/Newspaper': 'newspaper',
    'http://purl.org/spar/fabio/NewspaperArticle': 'newspaper article',
    'http://purl.org/spar/fabio/NewspaperIssue': 'newspaper issue',
    'http://purl.org/spar/fr/ReviewVersion': 'peer review', 
    'http://purl.org/spar/fabio/AcademicProceedings': 'proceedings',
    'http://purl.org/spar/fabio/Preprint': 'preprint',
    'http://purl.org/spar/fabio/Presentation': 'presentation',
    'http://purl.org/spar/fabio/ProceedingsPaper': 'proceedings article', 
    'http://purl.org/spar/fabio/ReferenceBook': 'reference book', 
    'http://purl.org/spar/fabio/ReferenceEntry': 'reference entry', 
    'http://purl.org/spar/fabio/ReportDocument': 'report', 
    'http://purl.org/spar/fabio/RetractionNotice': 'retraction notice',
    'http://purl.org/spar/fabio/Series': 'series', 
    'http://purl.org/spar/fabio/SpecificationDocument': 'standard', 
    'http://purl.org/spar/fabio/WebContent': 'web content'}

FIELDNAMES = ['id', 'title', 'author', 'issue', 'volume', 'venue', 'page', 'pub_date', 'type', 'publisher', 'editor']

def generate_csv(filename, meta_config: str, rdf_dir, dir_split_number, items_per_file, resp_agent, output_dir: str, zip_output_rdf: bool, depth=10) -> None:
    memory = dict()
    csv_data = get_csv_data(os.path.join(output_dir, filename))
    output_data = list()
    for row in csv_data:
        br_exists = True
        output_row = {
            'title': row['title'],
            'type': row['type'],
            'pub_date': row['pub_date']
        }
        row_identifiers = []
        row_omid = [identifier for identifier in row['id'].split() if identifier.startswith('omid')][0]
        row_identifiers.append(row_omid)
        for identifier in [identifier for identifier in row['id'].split() if not identifier.startswith('omid')]:
            id_path = find_file(rdf_dir, dir_split_number, items_per_file, identifier, zip_output_rdf)
            if id_path:
                id_info = process_archive(id_path, process_id, memory, depth, identifier, meta_config, resp_agent, id_path, memory, row_omid, depth)
                if isinstance(id_info, dict):
                    if 'error' in id_info:
                        if 'br does not exist' in id_info['error']:
                            br_exists = False
                            break
                        elif 'Maximum recursion depth reached' in id_info['error']:
                            continue
                row_identifiers.append(id_info)
            else:
                row_identifiers.append(identifier)
        if not br_exists:
            continue
        output_row['id'] = ' '.join(row_identifiers)
        agents_by_role = {'author': dict(), 'editor': dict(), 'publisher': dict()}
        last_roles = {'author': {'all': dict(), 'last': []}, 'editor': {'all': dict(), 'last': []}, 'publisher': {'all': dict(), 'last': []}}
        self_next = {'author': False, 'editor': False, 'publisher': False}
        is_cache = True
        for agent in row['author'].split('; '):
            agent_path = find_file(rdf_dir, dir_split_number, items_per_file, agent, zip_output_rdf)
            if agent_path:
                agent_info = process_archive(agent_path, process_agent, memory, depth, agent, meta_config, resp_agent, agent_path, memory, row_omid, depth)
                if isinstance(agent_info, dict):
                    if 'error' in agent_info:
                        if 'br does not exist' in agent_info['error']:
                            br_exists = False
                            break
                        elif 'Maximum recursion depth reached' in agent_info['error']:
                            continue
                agent_role = agent_info['role']
                if not agent_info['next']:
                    last_roles[agent_role]['last'].append(agent)
                if agent_info['next'] == agent:
                    self_next[agent_role] = True
                last_roles[agent_role]['all'][agent] = agent_info['next']
                agents_by_role[agent_role][agent] = agent_info
                is_cache = False
        if not br_exists:
            continue
        if not agents_by_role['author'] and not is_cache:
            pass
        # if not is_cache:
        #     agents_by_role = fix_roles(last_roles, self_next, meta_config, resp_agent, agents_by_role)
        for agent_role, agents in agents_by_role.items():
            last = ''
            new_role_list = list()
            iteration_count = 0
            while agents:
                iteration_count += 1
                if iteration_count > len(agents)*10: # è la soglia oltre la quale assumiamo ci sia un ciclo infinito
                    # Riordina gli agenti in ordine alfabetico
                    sorted_agents = sorted(agents.items(), key=lambda x: x[1]['ra'])
                    for i, (agent, data) in enumerate(sorted_agents[:-1]):
                        sorted_agents[i][1]['next'] = sorted_agents[i + 1][0]  # Aggiorna 'next' al successivo in lista
                    sorted_agents[-1][1]['next'] = ''  # L'ultimo elemento non ha un 'next'
                    # Prepara una nuova lista di agenti basata sull'ordine aggiornato
                    agents = dict(sorted_agents)
                    new_role_list.clear()
                    last = ''
                    iteration_count = 0  # Reset del contatore per la nuova iterazione
                for agent, agent_data in agents.items():
                    if agent_data['next'] == last:
                        new_role_list.append(agent_data['ra'])
                        last = agent
                        del agents[agent]
                        break
            if new_role_list:
                row[agent_role] = '; '.join(reversed(new_role_list))
        for role in ['author', 'editor', 'publisher']:
            for ra in row[role].split('; '):
                if ra:
                    ra_path = find_file(rdf_dir, dir_split_number, items_per_file, ra, zip_output_rdf)
                    if ra_path:
                        output_ra = process_archive(ra_path, process_responsible_agent, memory, depth, ra, rdf_dir, dir_split_number, items_per_file, memory, ra_path, meta_config, resp_agent, zip_output_rdf, row_omid, depth)
                        if isinstance(output_ra, dict):
                            if 'error' in output_ra:
                                if 'br does not exist' in output_ra['error']:
                                    br_exists = False
                                    break
                                elif 'Maximum recursion depth reached' in output_ra['error']:
                                    row[role] = row[role].replace(ra, '')
                                    output_row[role] = row[role]
                                    continue
                        row[role] = row[role].replace(ra, output_ra)
                        output_row[role] = row[role]
                    else:
                        output_row[role] = row[role]
        if 'omid:br/0643022790' in output_row['id']:
            print(output_row)
        if not br_exists:
            continue
        for venue in row['venue'].split():
            venue_path = find_file(rdf_dir, dir_split_number, items_per_file, venue, zip_output_rdf)
            if venue_path:
                to_be_found = venue
                while to_be_found:
                    venue_info, to_be_found = process_archive(venue_path, process_venue, memory, depth, to_be_found, rdf_dir, dir_split_number, items_per_file, venue_path, meta_config, resp_agent, memory, zip_output_rdf, row_omid, depth)
                    for k, v in venue_info.items():
                        if v:
                            output_row[k] = v
                    if to_be_found:
                        venue_path = find_file(rdf_dir, dir_split_number, items_per_file, to_be_found, zip_output_rdf)
            else:
                output_row['issue'] = row['issue']
                output_row['volume'] = row['volume']
                output_row['venue'] = row['venue']
        if row['page']:
            page_uri = row['page']
            page_path = find_file(rdf_dir, dir_split_number, items_per_file, page_uri, zip_output_rdf)
            if page_path:
                output_row['page'] = process_archive(page_path, process_page, memory, 2, page_uri, meta_config, resp_agent)
            else:
                output_row['page'] = row['page']
        output_data.append(output_row)
    write_csv(os.path.join(output_dir, filename), output_data, FIELDNAMES)

def process_archives(rdf_dir: str, entity_abbr:str, output_dir: str|None, doing_what: callable, dir_split_number: int, items_per_file: int, zip_output_rdf: bool, meta_config: str, resp_agent: str, wanted_schemas: list, memory: dict, threshold: int=3000, depth=2):
    br_files = [os.path.join(fold, file) for fold, _, files in os.walk(os.path.join(rdf_dir, entity_abbr)) for file in files if file.endswith('.zip') and os.path.basename(fold) != 'prov']
    print('[csv_generator: INFO] Looking for bibliographic resources recursively')
    pbar = tqdm(total=len(br_files))
    counter = 0
    global_output = list()
    for dirpath, _, filenames in os.walk(os.path.join(rdf_dir, entity_abbr)):
        memory = dict()
        for filename in filenames:
            if (filename.endswith('.zip') or filename.endswith('.json')) and os.path.basename(dirpath) != 'prov':
                local_output = process_archive(os.path.join(dirpath, filename), doing_what, memory, depth, rdf_dir, dir_split_number, items_per_file, zip_output_rdf, meta_config, resp_agent, wanted_schemas, memory, depth)
                global_output.extend(local_output)
                if len(global_output) > threshold and output_dir:
                    write_csv(os.path.join(output_dir, f'{counter}.csv'), global_output, FIELDNAMES)
                    counter += 1
                    global_output = list()
                pbar.update()
    if output_dir:
        write_csv(os.path.join(output_dir, f'{counter}.csv'), global_output, FIELDNAMES)
    pbar.close()

def process_archive(filepath: str, doing_what: callable, memory: dict|None, depth: int, *args) -> list:
    if depth == 0:
        return {'error': 'Maximum recursion depth reached'}

    filelock = FileLock(f'{filepath}.lock')
    if memory is not None:
        if filepath in memory:
            return doing_what(memory[filepath], *args)
    with filelock:
        try:
            if filepath.endswith('.zip'):
                with ZipFile(file=filepath, mode="r") as archive:
                    for zf_name in archive.namelist():
                        f = archive.open(zf_name)
            elif filepath.endswith('.json'):
                f = open(file=filepath, mode="r", encoding="utf8")
            data = json.load(f)
            f.close()
        except Exception as e:
            print('\nThe invalid zip file is', filepath, '\n', e)
            data = dict()
    if memory is not None:
        memory[filepath] = data
    return doing_what(data, *args)

def process_br(br_data: list, rdf_dir: str, dir_split_number: str, items_per_file: str, zip_output_rdf: bool, meta_config: str, resp_agent: str, wanted_schemas: list, memory: dict, depth: int) -> list:
    csv_br_data = list()
    for graph in br_data:
        graph_data = graph['@graph']
        for br in graph_data:
            row = dict()
            try:
                br_types = [x for x in br['@type'] if x != 'http://purl.org/spar/fabio/Expression']
            except KeyError:
                continue
            br_type = URI_TYPE_DICT[br_types[0]] if len(br_types) == 1 else ''
            if br_type in {'journal volume', 'journal issue'}:
                continue
            br_omid = f"omid:{br['@id'].split('/meta/')[1]}"
            br_identifiers = [br_omid]
            br_id_schemas = []
            if 'http://purl.org/spar/datacite/hasIdentifier' in br:
                br_ids = br['http://purl.org/spar/datacite/hasIdentifier']
                for br_identifier in br_ids:
                    if wanted_schemas is not None:
                        id_uri = br_identifier['@id']
                        id_path = find_file(rdf_dir, dir_split_number, items_per_file, id_uri, zip_output_rdf)
                        if id_path:
                            id_info = process_archive(id_path, process_id, memory, 2, id_uri, meta_config, resp_agent, id_path, memory, None, depth)
                            br_id_schemas.append(id_info.split(':', maxsplit=1)[0])
                    br_identifiers.append(br_identifier['@id'])
            if wanted_schemas is not None:
                if not set(wanted_schemas).intersection(br_id_schemas):
                    continue
            row['id'] = ' '.join(br_identifiers)
            row['title'] = br['http://purl.org/dc/terms/title'][0]['@value'] if 'http://purl.org/dc/terms/title' in br else ''
            row['type'] = br_type
            row['pub_date'] = br['http://prismstandard.org/namespaces/basic/2.0/publicationDate'][0]['@value'] if 'http://prismstandard.org/namespaces/basic/2.0/publicationDate' in br else ''
            br_ars = list()
            if 'http://purl.org/spar/pro/isDocumentContextFor' in br:
                for ar_data in br['http://purl.org/spar/pro/isDocumentContextFor']:
                    br_ars.append(ar_data['@id'])
            row['author'] = '; '.join(br_ars)
            row['page'] = br['http://purl.org/vocab/frbr/core#embodiment'][0]['@id'] if 'http://purl.org/vocab/frbr/core#embodiment' in br else ''
            if 'http://purl.org/vocab/frbr/core#partOf' in br:
                row['venue'] = br['http://purl.org/vocab/frbr/core#partOf'][0]['@id']
            csv_br_data.append(row)
    return csv_br_data

def process_id(id_data: list, id_uri: str, meta_config: str, resp_agent: str, id_path: str, memory: dict|None, br_uri: str, depth: int) -> str:
    output_identifier = None
    for graph in id_data:
        graph_data = graph['@graph']
        for id_data in graph_data:
            if id_data['@id'] == id_uri:
                id_schema = id_data['http://purl.org/spar/datacite/usesIdentifierScheme'][0]['@id'].split('/datacite/')[1]
                literal_value = id_data['http://www.essepuntato.it/2010/06/literalreification/hasLiteralValue'][0]['@value']
                output_identifier = f'{id_schema}:{literal_value}'
    if output_identifier:
        return output_identifier
    else:
        if memory is not None:
            del memory[id_path]
        meta_editor = MetaEditor(meta_config, resp_agent)
        sync_success = meta_editor.sync_rdf_with_triplestore(id_uri, source_uri=br_uri)
        if sync_success:
            return process_archive(id_path, process_id, memory, depth - 1, id_uri, meta_config, resp_agent, id_path, memory, br_uri, depth-1)
        else:
            return {'error': 'br does not exist'}

def process_agent(agent_data: list, agent_uri: str, meta_config: str, resp_agent: str, agent_path: str, memory: dict, row_omid: str, depth: int) -> dict:
    agent_dict = dict()
    for graph in agent_data:
        graph_data = graph['@graph']
        for agent in graph_data:
            if agent['@id'] == agent_uri:
                agent_dict['ra'] = agent['http://purl.org/spar/pro/isHeldBy'][0]['@id']
                agent_dict['role'] = agent['http://purl.org/spar/pro/withRole'][0]['@id'].split('/pro/')[1]
                agent_dict['next'] = ''
                if 'https://w3id.org/oc/ontology/hasNext' in agent:
                    agent_dict['next'] = agent['https://w3id.org/oc/ontology/hasNext'][0]['@id']
    if agent_dict:
        return agent_dict
    else:
        del memory[agent_path]
        meta_editor = MetaEditor(meta_config, resp_agent)
        br_uri = 'https://w3id.org/oc/meta/' + row_omid.replace('omid:', '')
        sync_success = meta_editor.sync_rdf_with_triplestore(agent_uri, br_uri)
        if sync_success:
            return process_archive(agent_path, process_agent, memory, depth-1, agent_uri, meta_config, resp_agent, agent_path, memory, br_uri, depth-1)
        else:
            return {'error': 'br does not exist'}
                
                    
def process_responsible_agent(ra_data: list, ra_uri: str, rdf_dir: str, dir_split_number: str, items_per_file: str, memory: dict, ra_path: str, meta_config: str, resp_agent: str, zip_output_rdf: bool, row_omid: str, depth: int) -> str:
    ra_value = None
    br_uri = 'https://w3id.org/oc/meta/' + row_omid.replace('omid:', '')
    for graph in ra_data:
        graph_data = graph['@graph']
        for ra in graph_data:
            if ra['@id'] == ra_uri:
                ra_name = ra['http://xmlns.com/foaf/0.1/name'][0]['@value'] if 'http://xmlns.com/foaf/0.1/name' in ra else None
                ra_fn = ra['http://xmlns.com/foaf/0.1/familyName'][0]['@value'] if 'http://xmlns.com/foaf/0.1/familyName' in ra else None
                ra_gn = ra['http://xmlns.com/foaf/0.1/givenName'][0]['@value'] if 'http://xmlns.com/foaf/0.1/givenName' in ra else None
                full_name = ''
                if ra_name:
                    full_name = ra_name
                else:
                    if ra_fn:
                        full_name += f'{ra_fn},'
                        if ra_gn:
                            full_name += f' {ra_gn}'
                    elif ra_gn:
                        full_name += f', {ra_gn}'
                ra_ids = [f"omid:{ra_uri.split('/meta/')[1]}"]
                if 'http://purl.org/spar/datacite/hasIdentifier' in ra:
                    for ra_identifier in ra['http://purl.org/spar/datacite/hasIdentifier']:
                        id_uri = ra_identifier['@id']
                        id_path = find_file(rdf_dir, dir_split_number, items_per_file, id_uri, zip_output_rdf)
                        id_info = process_archive(id_path, process_id, memory, 2, id_uri, meta_config, resp_agent, id_path, memory, br_uri, 2)
                        if isinstance(id_info, dict):
                            if 'error' in id_info:
                                if 'br does not exist' in id_info['error']:
                                    return {'error': 'br does not exist'}
                                elif 'Maximum recursion depth reached' in id_info['error']:
                                    continue
                        ra_ids.append(id_info)
                if full_name:
                    ra_value = f"{full_name} [{' '.join(ra_ids)}]"
                else:
                    ra_value = f"[{' '.join(ra_ids)}]"
    if ra_value:
        return ra_value
    else:
        del memory[ra_path]
        meta_editor = MetaEditor(meta_config, resp_agent)
        sync_success = meta_editor.sync_rdf_with_triplestore(ra_uri, source_uri=br_uri)
        if sync_success:
            return process_archive(ra_path, process_responsible_agent, memory, depth-1, ra_uri, rdf_dir, dir_split_number, items_per_file, memory, ra_path, meta_config, resp_agent, zip_output_rdf, row_omid, depth-1)
        else:
            return {'error': 'br does not exist'}

def process_venue(venue_data: list, venue_uri: str, rdf_dir: str, dir_split_number: str, items_per_file: str, venue_path: str, meta_config: str, resp_agent: str, memory: dict, zip_output_rdf: bool, br_uri: str, depth: int) -> Tuple[dict, list]:
    venue_dict = {'volume': '', 'issue': '', 'venue': ''}
    to_be_found = None
    for graph in venue_data:
        graph_data = graph['@graph']
        for venue in graph_data:
            if venue['@id'] == venue_uri:
                if 'http://purl.org/spar/fabio/JournalIssue' in venue['@type']:
                    venue_dict['issue'] = venue['http://purl.org/spar/fabio/hasSequenceIdentifier'][0]['@value']
                elif 'http://purl.org/spar/fabio/JournalVolume' in venue['@type']:
                    venue_dict['volume'] = venue['http://purl.org/spar/fabio/hasSequenceIdentifier'][0]['@value']
                else:
                    venue_title = venue['http://purl.org/dc/terms/title'][0]['@value'] if 'http://purl.org/dc/terms/title' in venue else None
                    venue_ids = venue['http://purl.org/spar/datacite/hasIdentifier'] if 'http://purl.org/spar/datacite/hasIdentifier' in venue else []
                    explicit_ids = [f"omid:{venue_uri.split('/meta/')[1]}"]
                    for venue_id in venue_ids:
                        id_path = find_file(rdf_dir, dir_split_number, items_per_file, venue_id['@id'], zip_output_rdf)
                        explicit_ids.append(process_archive(id_path, process_id, memory, 2, venue_id['@id'], meta_config, resp_agent, id_path, memory, br_uri, 2))
                    venue_full = f"{venue_title} [{' '.join(explicit_ids)}]" if venue_title else f"[{' '.join(explicit_ids)}]"
                    venue_dict['venue'] = venue_full
                if 'http://purl.org/vocab/frbr/core#partOf' in venue:
                    venue_partOf = venue['http://purl.org/vocab/frbr/core#partOf'][0]['@id']
                    if venue_uri == venue_partOf:
                        meta_editor = MetaEditor(meta_config, resp_agent)
                        meta_editor.delete(res=venue_uri, property='is_part_of')
                    else:
                        to_be_found = venue_partOf
    if any(v for _, v in venue_dict.items()):
        return venue_dict, to_be_found
    else:
        del memory[venue_path]
        meta_editor = MetaEditor(meta_config, resp_agent)
        meta_editor.sync_rdf_with_triplestore(venue_uri)
        return process_archive(venue_path, process_venue, memory, depth-1, venue_uri, rdf_dir, dir_split_number, items_per_file, venue_path, meta_config, resp_agent, memory, zip_output_rdf, br_uri, depth-1)

def process_page(page_data: list, page_uri: str, meta_config: str, resp_agent: str) -> str:
    for graph in page_data:
        graph_data = graph['@graph']
        for venue in graph_data:
            if venue['@id'] == page_uri:
                starting_page = None
                ending_page = None
                if 'http://prismstandard.org/namespaces/basic/2.0/startingPage' in venue:
                    starting_page = venue['http://prismstandard.org/namespaces/basic/2.0/startingPage'][0]['@value']
                if 'http://prismstandard.org/namespaces/basic/2.0/endingPage' in venue:
                    ending_page = venue['http://prismstandard.org/namespaces/basic/2.0/endingPage'][0]['@value']
                if not all([starting_page, ending_page]):
                    meta_editor = MetaEditor(meta_config, resp_agent)
                    if not starting_page and ending_page:
                        meta_editor.update_property(URIRef(page_uri), 'has_starting_page', ending_page)
                        starting_page = ending_page
                    elif not ending_page and starting_page:
                        meta_editor.update_property(URIRef(page_uri), 'has_ending_page', starting_page)
                        ending_page = starting_page
                    else:
                        # TODO: delete entity
                        return ''
                return f'{starting_page}-{ending_page}'

def find_file(rdf_dir: str, dir_split_number: str, items_per_file: str, uri: str, zip_output_rdf: bool) -> str|None:
    entity_regex: str = r'^(https:\/\/w3id\.org\/oc\/meta)\/([a-z][a-z])\/(0[1-9]+0)?([1-9][0-9]*)$'
    entity_match = re.match(entity_regex, uri)
    if entity_match:
        cur_number = int(entity_match.group(4))
        cur_file_split: int = 0
        while True:
            if cur_number > cur_file_split:
                cur_file_split += items_per_file
            else:
                break
        cur_split: int = 0
        while True:
            if cur_number > cur_split:
                cur_split += dir_split_number
            else:
                break
        short_name = entity_match.group(2)
        sub_folder = entity_match.group(3)
        cur_dir_path = os.path.join(rdf_dir, short_name, sub_folder, str(cur_split))
        extension = '.zip' if zip_output_rdf else '.json'
        cur_file_path = os.path.join(cur_dir_path, str(cur_file_split)) + extension
        return cur_file_path

def fix_roles(last_roles: dict, self_next: dict, meta_config: str, resp_agent: str, agents_by_role: dict) -> dict:
    meta_editor = MetaEditor(meta_config, resp_agent)
    for role_type, role_data in last_roles.items():
        all_list = list(role_data['all'].keys())
        last_list = role_data['last']
        if (all_list and len(last_list) != 1) or self_next[role_type]:
            sorted_roles_list = sorted(all_list)
            for i, role in enumerate(sorted_roles_list):
                if i < len(sorted_roles_list) - 1:
                    if last_roles[role_type]['all'][role] != sorted_roles_list[i+1]:
                        meta_editor.delete(URIRef(role), 'has_next')
                elif i == len(sorted_roles_list) - 1:
                    if last_roles[role_type]['all'][role]:
                        meta_editor.delete(URIRef(role), 'has_next')
            for i, role in enumerate(sorted_roles_list):
                if i < len(sorted_roles_list) - 1:
                    if last_roles[role_type]['all'][role] != sorted_roles_list[i+1]:
                        meta_editor.update_property(URIRef(role), 'has_next', URIRef(sorted_roles_list[i+1]))
            for agent_role, _ in agents_by_role.items():
                if agent_role == role_type:
                    new_agents = {sorted_role: {'next': sorted_roles_list[i+1], 'ra': agents_by_role[agent_role][sorted_role]['ra'], 'role': agents_by_role[agent_role][sorted_role]['role']} if i < len(sorted_roles_list) - 1 else {'next': '', 'ra': agents_by_role[agent_role][sorted_role]['ra'], 'role': agents_by_role[agent_role][sorted_role]['role']} for i, sorted_role in enumerate(sorted_roles_list)}
                    agents_by_role[agent_role] = new_agents
    return agents_by_role


#!python
# Copyright 2023, Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


from argparse import ArgumentParser

from oc_ocdm.graph.graph_entity import GraphEntity
from pebble import ProcessFuture, ProcessPool
from SPARQLWrapper import JSON, SPARQLWrapper
from tqdm import tqdm

from oc_meta.lib.csvmanager import CSVManager


def task_done(task_output:ProcessFuture) -> None:
    identifier, it_exists = task_output.result()
    if not it_exists:
        print(identifier)
    PBAR.update()

def ask_for_id(identifier: str, sparql: SPARQLWrapper) -> bool:
    query = f'''
        ASK {{
            ?br <{GraphEntity.iri_has_identifier}> ?identifier.
            ?identifier <{GraphEntity.iri_uses_identifier_scheme}> <{GraphEntity.iri_doi}>;
                        <{GraphEntity.iri_has_literal_value}> "{identifier}".
        }}
    '''
    sparql.setQuery(query)
    it_exists = sparql.queryAndConvert()['boolean']
    return identifier, it_exists

if __name__ == '__main__':
    arg_parser = ArgumentParser('check_meta_output.py')
    arg_parser.add_argument('-i', '--ids', dest='ids', required=False,
                            help='A CSV filepath containing what ids must be checked')
    arg_parser.add_argument('-t', '--ts', dest='ts', required=False,
                            help='The triplestore endpoint URL')
    args = arg_parser.parse_args()
    ids_set = list(CSVManager.load_csv_column_as_set(args.ids, 'id'))
    PBAR = tqdm(total=len(ids_set))
    print('Starting the queries')
    sparql = SPARQLWrapper(args.ts)
    sparql.setReturnFormat(JSON)
    with ProcessPool() as executor:
        for identifier in ids_set:
            future:ProcessFuture = executor.schedule(
                function=ask_for_id, 
                args=(identifier, sparql)) 
            future.add_done_callback(task_done)
    PBAR.close()




#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2019 Silvio Peroni <essepuntato@gmail.com>
# Copyright 2019-2020 Fabio Mariani <fabio.mariani555@gmail.com>
# Copyright 2021 Simone Persiani <iosonopersia@gmail.com>
# Copyright 2021-2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from __future__ import annotations

import csv
import os
import traceback
from argparse import ArgumentParser
from concurrent.futures import as_completed
from datetime import datetime
from itertools import cycle
from sys import executable, platform
from typing import List, Tuple

import redis
import yaml
from oc_meta.core.creator import Creator
from oc_meta.core.curator import Curator
from oc_meta.lib.file_manager import (get_csv_data, init_cache, normalize_path,
                                      pathoo, sort_files, zipit)
from oc_meta.plugins.multiprocess.resp_agents_creator import RespAgentsCreator
from oc_meta.plugins.multiprocess.resp_agents_curator import RespAgentsCurator
from oc_meta.run.upload.on_triplestore import *
from pebble import ProcessPool
from time_agnostic_library.support import generate_config_file
from tqdm import tqdm

from oc_ocdm import Storer
from oc_ocdm.counter_handler.redis_counter_handler import RedisCounterHandler
from oc_ocdm.prov import ProvSet
from oc_ocdm.support.reporter import Reporter


class MetaProcess:
    def __init__(self, settings:dict, meta_config_path: str):
        # Mandatory settings
        self.triplestore_url = settings['triplestore_url']
        self.input_csv_dir = normalize_path(settings['input_csv_dir'])
        self.base_output_dir = normalize_path(settings['base_output_dir'])
        self.resp_agent = settings['resp_agent']
        self.info_dir = os.path.join(self.base_output_dir, 'info_dir')
        self.output_csv_dir = os.path.join(self.base_output_dir, 'csv')
        self.distinct_output_dirs = True if settings['base_output_dir'] != settings['output_rdf_dir'] else False
        self.output_rdf_dir = normalize_path(settings['output_rdf_dir']) + os.sep + 'rdf' + os.sep
        self.indexes_dir = os.path.join(self.base_output_dir, 'indexes')
        self.cache_path = os.path.join(self.base_output_dir, 'cache.txt')
        self.errors_path = os.path.join(self.base_output_dir, 'errors.txt')
        # Optional settings
        self.base_iri = settings['base_iri']
        self.normalize_titles = settings.get('normalize_titles', True)
        self.context_path = settings['context_path']
        self.dir_split_number = settings['dir_split_number']
        self.items_per_file = settings['items_per_file']
        self.default_dir = settings['default_dir']
        self.rdf_output_in_chunks = settings['rdf_output_in_chunks']
        self.zip_output_rdf = settings['zip_output_rdf']
        self.source = settings['source']
        self.valid_dois_cache = dict() if bool(settings['use_doi_api_service']) == True else None
        self.workers_number = int(settings['workers_number'])
        supplier_prefix:str = settings['supplier_prefix']
        self.supplier_prefix = supplier_prefix[:-1] if supplier_prefix.endswith('0') else supplier_prefix
        self.silencer = settings['silencer']
        # Time-Agnostic_library integration
        self.time_agnostic_library_config = os.path.join(os.path.dirname(meta_config_path), 'time_agnostic_library_config.json')
        if not os.path.exists(self.time_agnostic_library_config):
            generate_config_file(config_path=self.time_agnostic_library_config, dataset_urls=[self.triplestore_url], dataset_dirs=list(),
                provenance_urls=settings['provenance_endpoints'], provenance_dirs=list(), 
                blazegraph_full_text_search=settings['blazegraph_full_text_search'], fuseki_full_text_search=settings['fuseki_full_text_search'], 
                virtuoso_full_text_search=settings['virtuoso_full_text_search'], graphdb_connector_name=settings['graphdb_connector_name'], 
                cache_endpoint=settings['cache_endpoint'], cache_update_endpoint=settings['cache_update_endpoint'])

        # Redis settings
        self.redis_host = settings.get('redis_host', 'localhost')
        self.redis_port = settings.get('redis_port', 6379)
        self.redis_db = settings.get('redis_db', 5)
        self.redis_client = redis.Redis(host=self.redis_host, port=self.redis_port, db=self.redis_db)

        self.counter_handler = RedisCounterHandler(host=self.redis_host, port=self.redis_port, db=self.redis_db)

    def prepare_folders(self) -> List[str]:
        completed = init_cache(self.cache_path)
        files_in_input_csv_dir = {filename for filename in os.listdir(self.input_csv_dir) if filename.endswith('.csv')}
        files_to_be_processed = sort_files(list(files_in_input_csv_dir.difference(completed)))
        for dir in [self.output_csv_dir, self.indexes_dir, self.output_rdf_dir]:
            pathoo(dir)
        if self.rdf_output_in_chunks:
            data_dir = os.path.join(self.output_rdf_dir, 'data' + os.sep)
            prov_dir = os.path.join(self.output_rdf_dir, 'prov' + os.sep)
            for dir in [prov_dir, data_dir]:
                pathoo(dir)
        csv.field_size_limit(128)
        return files_to_be_processed

    def curate_and_create(self, filename:str, cache_path:str, errors_path:str, worker_number:int=None, resp_agents_only:bool=False, settings:str|None=None, meta_config_path: str = None) -> Tuple[dict, str, str, str]:
        if os.path.exists(os.path.join(self.base_output_dir, '.stop')):
            return {'message': 'skip'}, cache_path, errors_path, filename
        try:
            filepath = os.path.join(self.input_csv_dir, filename)
            print(filepath)
            data = get_csv_data(filepath)
            supplier_prefix = f'{self.supplier_prefix}0' if worker_number is None else f'{self.supplier_prefix}{str(worker_number)}0'
            # Curator
            self.info_dir = os.path.join(self.info_dir, supplier_prefix)
            curator_info_dir = os.path.join(self.info_dir, 'curator' + os.sep)
            if resp_agents_only:
                curator_obj = RespAgentsCurator(
                    data=data, 
                    ts=self.triplestore_url, 
                    prov_config=self.time_agnostic_library_config, 
                    counter_handler=self.counter_handler,
                    base_iri=self.base_iri, 
                    prefix=supplier_prefix, 
                    settings=settings,
                    meta_config_path=meta_config_path)
            else:
                curator_obj = Curator(
                    data=data, 
                    ts=self.triplestore_url, 
                    prov_config=self.time_agnostic_library_config, 
                    counter_handler=self.counter_handler,
                    base_iri=self.base_iri, 
                    prefix=supplier_prefix, 
                    valid_dois_cache=self.valid_dois_cache, 
                    settings=settings,
                    silencer=self.silencer,
                    meta_config_path=meta_config_path
                )
            name = f"{filename.replace('.csv', '')}_{datetime.now().strftime('%Y-%m-%dT%H-%M-%S')}"
            curator_obj.curator(filename=name, path_csv=self.output_csv_dir, path_index=self.indexes_dir)
            # Creator
            if resp_agents_only:
                creator_obj = RespAgentsCreator(
                    data=curator_obj.data, 
                    endpoint=self.triplestore_url, 
                    base_iri=self.base_iri,
                    counter_handler=self.counter_handler,
                    supplier_prefix=supplier_prefix, 
                    resp_agent=self.resp_agent, 
                    ra_index=curator_obj.index_id_ra,
                    preexisting_entities=curator_obj.preexisting_entities, 
                    everything_everywhere_allatonce=curator_obj.everything_everywhere_allatonce, 
                    settings=settings,
                    meta_config_path=meta_config_path)
            else:
                creator_obj = Creator(
                    data=curator_obj.data, 
                    endpoint=self.triplestore_url, 
                    base_iri=self.base_iri, 
                    counter_handler=self.counter_handler,
                    supplier_prefix=supplier_prefix, 
                    resp_agent=self.resp_agent, 
                    ra_index=curator_obj.index_id_ra,
                    br_index=curator_obj.index_id_br, 
                    re_index_csv=curator_obj.re_index, 
                    ar_index_csv=curator_obj.ar_index, 
                    vi_index=curator_obj.VolIss, 
                    preexisting_entities=curator_obj.preexisting_entities, 
                    everything_everywhere_allatonce=curator_obj.everything_everywhere_allatonce, 
                    settings=settings,
                    meta_config_path=meta_config_path
                )
            creator = creator_obj.creator(source=self.source)
            # Provenance
            prov = ProvSet(creator, self.base_iri, wanted_label=False, supplier_prefix=supplier_prefix, custom_counter_handler=self.counter_handler)
            modified_entities = prov.generate_provenance()
            # Storer
            repok  = Reporter(print_sentences = False)
            reperr = Reporter(print_sentences = True, prefix='[Storer: ERROR] ')
            res_storer = Storer(abstract_set=creator, repok=repok, reperr=reperr, context_map={}, dir_split=self.dir_split_number, n_file_item=self.items_per_file, default_dir=self.default_dir, output_format='json-ld', zip_output=self.zip_output_rdf, modified_entities=modified_entities)
            prov_storer = Storer(abstract_set=prov, repok=repok, reperr=reperr, context_map={}, dir_split=self.dir_split_number, n_file_item=self.items_per_file, output_format='json-ld', zip_output=self.zip_output_rdf, modified_entities=modified_entities)
            # with suppress_stdout():
            self.store_data_and_prov(res_storer, prov_storer, filename)
            return {'message': 'success'}, cache_path, errors_path, filename
        except Exception as e:
            tb = traceback.format_exc()
            template = "An exception of type {0} occurred. Arguments:\n{1!r}\nTraceback:\n{2}"
            message = template.format(type(e).__name__, e.args, tb)
            return {'message': message}, cache_path, errors_path, filename
    
    def store_data_and_prov(self, res_storer:Storer, prov_storer:Storer, filename:str) -> None:
        if self.rdf_output_in_chunks:
            filename_without_csv = filename[:-4]
            f = os.path.join(self.output_rdf_dir, 'data', filename_without_csv + '.json')
            res_storer.store_graphs_in_file(f, self.context_path)
            res_storer.upload_all(self.triplestore_url, self.output_rdf_dir, batch_size=10)
            f_prov = os.path.join(self.output_rdf_dir, 'prov', filename_without_csv + '.json')
            prov_storer.store_graphs_in_file(f_prov, self.context_path)
        else:
            res_storer.store_all(base_dir=self.output_rdf_dir, base_iri=self.base_iri, context_path=self.context_path, process_id=None)
            prov_storer.store_all(self.output_rdf_dir, self.base_iri, self.context_path, process_id=None)
            res_storer.upload_all(triplestore_url=self.triplestore_url, base_dir=self.output_rdf_dir, batch_size=10, save_queries=True)

    def save_data(self):
        output_dirname = f"meta_output_{datetime.now().strftime('%Y-%m-%dT%H_%M_%S_%f')}.zip"
        dirs_to_zip = [self.base_output_dir, self.output_rdf_dir] if self.distinct_output_dirs else [self.base_output_dir]
        zipit(dirs_to_zip, output_dirname)

    def run_sparql_updates(self, endpoint: str, folder: str, batch_size: int = 10):
        upload_sparql_updates(endpoint, folder, batch_size)

def run_meta_process(settings: dict, meta_config_path: str, resp_agents_only: bool=False) -> None:
    meta_process = MetaProcess(settings=settings, meta_config_path=meta_config_path)
    is_unix = platform in {'linux', 'linux2', 'darwin'}
    # delete_lock_files(base_dir=meta_process.base_output_dir)
    files_to_be_processed = meta_process.prepare_folders()
    max_workers = meta_process.workers_number
    if max_workers == 0:
        workers = list(range(1, os.cpu_count()))
    elif max_workers == 1:
        workers = [None]
    else:
        tens = int(str(max_workers)[:-1]) if max_workers >= 10 else 0
        multiples_of_ten = {i for i in range(1, max_workers + tens + 1) if int(i) % 10 == 0}
        workers = [i for i in range(1, max_workers+len(multiples_of_ten)+1) if i not in multiples_of_ten]
    generate_gentle_buttons(meta_process.base_output_dir, meta_config_path, is_unix)
    with ProcessPool(max_workers=max_workers, max_tasks=1) as executor, tqdm(total=len(files_to_be_processed), desc="Processing files") as progress_bar:
        futures = [executor.schedule(curate_and_create_wrapper, args=(file, worker, resp_agents_only, settings, meta_config_path)) 
                for file, worker in zip(files_to_be_processed, cycle(workers))]
        for future in as_completed(futures):
            try:
                result = future.result()
                task_done(result)  # Gestisci il risultato del task
            except Exception as e:
                # Gestisci l'eccezione per il task che ha sollevato un'errore
                traceback_str = traceback.format_exc()
                print(f"Errore durante l'elaborazione: {e}\nTraceback:\n{traceback_str}")
            finally:
                progress_bar.update(1)

    if not os.path.exists(os.path.join(meta_process.base_output_dir, '.stop')):
        if os.path.exists(meta_process.cache_path):
            os.rename(meta_process.cache_path, meta_process.cache_path.replace('.txt', f'_{datetime.now().strftime("%Y-%m-%dT%H_%M_%S_%f")}.txt'))
        if is_unix:
            delete_lock_files(base_dir=meta_process.base_output_dir)

    meta_process.run_sparql_updates(endpoint=settings['triplestore_url'], folder=os.path.join(meta_process.output_rdf_dir, 'to_be_uploaded'))

def curate_and_create_wrapper(file_to_be_processed, worker_number, resp_agents_only, settings, meta_config_path):
    meta_process = MetaProcess(settings=settings, meta_config_path=meta_config_path)
    return meta_process.curate_and_create(file_to_be_processed, meta_process.cache_path, meta_process.errors_path, worker_number, resp_agents_only, settings, meta_config_path)

def task_done(task_output: tuple) -> None:
    message, cache_path, errors_path, filename = task_output
    if message['message'] == 'skip':
        pass
    elif message['message'] == 'success':
        if not os.path.exists(cache_path):
            with open(cache_path, 'w', encoding='utf-8') as aux_file:
                aux_file.write(filename + '\n')
        else:
            with open(cache_path, 'r', encoding='utf-8') as aux_file:
                cache_data = aux_file.read().splitlines()
                cache_data.append(filename)
                try:
                    data_sorted = sorted(cache_data, key=lambda filename: int(filename.replace('.csv', '')), reverse=False)
                except ValueError:
                    data_sorted = cache_data
            with open(cache_path, 'w', encoding='utf-8') as aux_file:
                aux_file.write('\n'.join(data_sorted))
    else:
        with open(errors_path, 'a', encoding='utf-8') as aux_file:
            aux_file.write(f'{filename}: {message["message"]}' + '\n')

def chunks(lst:list, n:int) -> List[list]:
    '''Yield successive n-sized chunks from lst.'''
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def delete_lock_files(base_dir:list) -> None:
    for dirpath, _, filenames in os.walk(base_dir):
        for filename in filenames:
            if filename.endswith('.lock'):
                os.remove(os.path.join(dirpath, filename))

def generate_gentle_buttons(dir:str, config:str, is_unix:bool):
    if os.path.exists(os.path.join(dir, '.stop')):
        os.remove(os.path.join(dir, '.stop'))
    ext = 'sh' if is_unix else 'bat'
    with open (f'gently_run.{ext}', 'w') as rsh:
        rsh.write(f'{executable} -m oc_meta.lib.stopper -t "{dir}" --remove\n{executable} -m oc_meta.run.meta_process -c {config}')
    with open (f'gently_stop.{ext}', 'w') as rsh:
        rsh.write(f'{executable} -m oc_meta.lib.stopper -t "{dir}" --add')


if __name__ == '__main__': # pragma: no cover
    arg_parser = ArgumentParser('meta_process.py', description='This script runs the OCMeta data processing workflow')
    arg_parser.add_argument('-c', '--config', dest='config', required=True, help='Configuration file directory')
    args = arg_parser.parse_args()
    with open(args.config, encoding='utf-8') as file:
        settings = yaml.full_load(file)
    run_meta_process(settings=settings, meta_config_path=args.config, resp_agents_only=False)


import argparse
import csv
import os

import pandas as pd
from retrying import retry
from SPARQLWrapper import JSON, SPARQLWrapper
from tqdm import tqdm


class UnionFind:
    def __init__(self):
        self.parent = {}

    def find(self, item):
        path = [item]
        while item in self.parent and self.parent[item] != item:
            item = self.parent[item]
            path.append(item)
            if len(path) > 1000:  # Limite arbitrario per evitare loop infiniti
                print(f"Warning: Long path detected: {' -> '.join(path)}")
                break
        for p in path:
            self.parent[p] = item
        return item

    def union(self, item1, item2):
        root1 = self.find(item1)
        root2 = self.find(item2)
        if root1 != root2:
            self.parent[root2] = root1


def load_csv(file_path):
    return pd.read_csv(file_path)


@retry(stop_max_attempt_number=3, wait_exponential_multiplier=1000, wait_exponential_max=10000)
def query_sparql(endpoint, uri, query_type):
    sparql = SPARQLWrapper(endpoint)
    
    if query_type == 'subjects':
        query = f"""
        SELECT ?subject WHERE {{
            ?subject ?predicate <{uri}> .
        }}
        """
    elif query_type == 'objects':
        query = f"""
        SELECT ?object WHERE {{
            <{uri}> ?predicate ?object .
            ?object ?p ?o .
        }}
        """
    
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    results = sparql.query().convert()
    
    if query_type == 'subjects':
        return [result['subject']['value'] for result in results['results']['bindings']]
    elif query_type == 'objects':
        return [result['object']['value'] for result in results['results']['bindings']]


def get_all_related_entities(endpoint, uris):
    related_entities = set(uris)
    for uri in uris:
        subjects = query_sparql(endpoint, uri, 'subjects')
        objects = query_sparql(endpoint, uri, 'objects')
        related_entities.update(subjects)
        related_entities.update(objects)
    return related_entities


def group_entities(df, endpoint):
    uf = UnionFind()
    
    for index, row in tqdm(df.iterrows(), total=df.shape[0], desc="Processing rows"):
        surviving_entity = row['surviving_entity']
        merged_entities = row['merged_entities'].split("; ")

        all_entities = [surviving_entity] + merged_entities
        
        all_related_entities = get_all_related_entities(endpoint, all_entities)
        
        for entity in all_related_entities:
            uf.union(surviving_entity, entity)

    grouped_data = {}
    for index, row in df.iterrows():
        surviving_entity = row['surviving_entity']
        group_id = uf.find(surviving_entity)
        
        if group_id not in grouped_data:
            grouped_data[group_id] = []
        
        grouped_data[group_id].append(row)

    # Converti le liste di righe in DataFrame
    for group_id in grouped_data:
        grouped_data[group_id] = pd.DataFrame(grouped_data[group_id])

    return grouped_data


def optimize_groups(grouped_data, target_size=50):
    """
    Ottimizza i gruppi combinando quelli singoli mantenendo separate le entità interconnesse.
    
    Args:
        grouped_data (dict): Dizionario di DataFrame raggruppati
        target_size (int): Dimensione minima target per ogni gruppo
    
    Returns:
        dict: Dizionario ottimizzato dei DataFrame raggruppati
    """
    # Separa i gruppi in singoli e multipli
    single_groups = {k: v for k, v in grouped_data.items() if len(v) == 1}
    multi_groups = {k: v for k, v in grouped_data.items() if len(v) > 1}
    
    # Se non ci sono gruppi singoli, restituisci i gruppi originali
    if not single_groups:
        return grouped_data
    
    # Crea nuovi gruppi combinando quelli singoli
    combined_groups = {}
    single_items = list(single_groups.items())
    
    # Combina i gruppi singoli in gruppi della dimensione target
    current_group = []
    current_key = None
    
    for key, df in single_items:
        if len(current_group) == 0:
            current_key = key
            
        current_group.append(df)
        
        if len(current_group) >= target_size:
            combined_groups[current_key] = pd.concat(current_group, ignore_index=True)
            current_group = []
    
    # Gestisci eventuali gruppi rimanenti
    if current_group:
        if len(current_group) == 1 and multi_groups:
            # Se è rimasto un gruppo singolo e ci sono gruppi multipli,
            # aggiungiamo il gruppo singolo al gruppo multiplo più piccolo
            smallest_multi = min(multi_groups.items(), key=lambda x: len(x[1]))
            multi_groups[smallest_multi[0]] = pd.concat(
                [smallest_multi[1]] + current_group, 
                ignore_index=True
            )
        else:
            # Altrimenti lo manteniamo come gruppo separato
            combined_groups[current_key] = pd.concat(current_group, ignore_index=True)
    
    # Unisci i gruppi multipli originali con i nuovi gruppi combinati
    return {**multi_groups, **combined_groups}


def save_grouped_entities(grouped_data, output_dir):
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
        
    for key, df in grouped_data.items():
        output_file = os.path.join(output_dir, f"{key.split('/')[-1]}.csv")
        print(f"Saving group with {len(df)} rows to {output_file}")
        
        try:
            df.to_csv(output_file, index=False)
        except AttributeError as e:
            print(f"Error saving file {output_file}: {str(e)}")
            try:
                df.to_csv(output_file, index=False, encoding='utf-8', quoting=csv.QUOTE_NONNUMERIC)
                print(f"Successfully saved using alternative method: {output_file}")
            except Exception as alt_e:
                print(f"Alternative method also failed: {str(alt_e)}")
        except Exception as e:
            print(f"Unexpected error saving file {output_file}: {str(e)}")


def main():
    parser = argparse.ArgumentParser(description='Process CSV and group entities based on SPARQL queries.')
    parser.add_argument('csv_file_path', type=str, help='Path to the input CSV file')
    parser.add_argument('output_dir', type=str, help='Directory to save the output files')
    parser.add_argument('sparql_endpoint', type=str, help='SPARQL endpoint URL')
    parser.add_argument('--min_group_size', type=int, default=50, 
                      help='Minimum target size for groups (default: 10)')

    args = parser.parse_args()
    
    df = load_csv(args.csv_file_path)
    print(f"Loaded CSV file with {len(df)} rows")
    
    grouped_entities = group_entities(df, args.sparql_endpoint)
    print(f"Initially grouped entities into {len(grouped_entities)} groups")
    
    optimized_groups = optimize_groups(grouped_entities, args.min_group_size)
    print(f"Optimized into {len(optimized_groups)} groups")
    
    save_grouped_entities(optimized_groups, args.output_dir)
    print("Finished saving grouped entities")


if __name__ == "__main__":
    main()


import argparse
import csv
import logging
import multiprocessing as mp
import os
import zipfile
from collections import defaultdict
from typing import Dict, Set

from rdflib import ConjunctiveGraph, URIRef
from tqdm import tqdm

logging.basicConfig(filename='error_log_find_duplicated_ids_from_files.txt', level=logging.ERROR, 
                    format='%(asctime)s - %(levelname)s - %(message)s')

def process_zip_file(zip_path: str) -> Dict[tuple, Set[str]]:
    entity_info = defaultdict(set)
    datacite_uses_identifier_scheme = URIRef("http://purl.org/spar/datacite/usesIdentifierScheme")
    literal_reification_has_literal_value = URIRef("http://www.essepuntato.it/2010/06/literalreification/hasLiteralValue")

    try:
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            for zip_file in zip_ref.namelist():
                try:
                    with zip_ref.open(zip_file) as rdf_file:
                        g = ConjunctiveGraph()
                        g.parse(data=rdf_file.read(), format="json-ld")
                        
                        for s, p, o in g.triples((None, datacite_uses_identifier_scheme, None)):
                            entity_id = str(s)
                            identifier_scheme = str(o)
                            literal_value = g.value(s, literal_reification_has_literal_value)
                            if identifier_scheme and literal_value:
                                key = (str(identifier_scheme), str(literal_value))
                                entity_info[key].add(entity_id)
                except Exception as e:
                    logging.error(f"Errore nell'elaborazione del file {zip_file} in {zip_path}: {str(e)}")
    except zipfile.BadZipFile:
        logging.error(f"File ZIP corrotto o non valido: {zip_path}")
    except Exception as e:
        logging.error(f"Errore nell'apertura del file ZIP {zip_path}: {str(e)}")

    return entity_info

def read_and_analyze_zip_files(folder_path: str, csv_path: str):
    id_folder_path = os.path.join(folder_path, 'id')

    if not os.path.exists(id_folder_path):
        logging.error(f"La sottocartella 'id' non esiste nel percorso: {folder_path}")
        return

    zip_files = [os.path.join(root, file) for root, _, files in os.walk(id_folder_path) 
                 for file in files if file.endswith('.zip') and file != 'se.zip']

    with mp.Pool(processes=mp.cpu_count()) as pool:
        results = list(tqdm(pool.imap(process_zip_file, zip_files), total=len(zip_files), desc="Analizzando i file ZIP"))

    entity_info = defaultdict(set)
    for result in results:
        for key, value in result.items():
            entity_info[key].update(value)

    save_duplicates_to_csv(entity_info, csv_path)

def save_duplicates_to_csv(entity_info: Dict[tuple, Set[str]], csv_path: str):
    try:
        with open(csv_path, mode='w', newline='', encoding='utf-8') as csv_file:
            csv_writer = csv.writer(csv_file)
            csv_writer.writerow(['surviving_entity', 'merged_entities'])

            for ids in entity_info.values():
                if len(ids) > 1:
                    ids_list = list(ids)
                    csv_writer.writerow([ids_list[0], '; '.join(ids_list[1:])])
    except Exception as e:
        logging.error(f"Errore nel salvataggio del file CSV {csv_path}: {str(e)}")

def main():
    parser = argparse.ArgumentParser(description="Legge i file RDF all'interno dei file ZIP in una sottocartella 'id'.")
    parser.add_argument("folder_path", type=str, help="Percorso della cartella contenente la sottocartella 'id'")
    parser.add_argument("csv_path", type=str, help="Percorso del file CSV per salvare i duplicati")
    args = parser.parse_args()

    read_and_analyze_zip_files(args.folder_path, args.csv_path)

if __name__ == "__main__":
    main()


import argparse
import multiprocessing
import os
import zipfile
from tqdm import tqdm
import json

def process_file(file_path):
    if file_path.endswith('.lock'):
        os.remove(file_path)
    elif file_path.endswith('.zip'):
        try:
            with zipfile.ZipFile(file_path, 'r') as zip_ref:
                with zipfile.ZipFile(file_path + '.tmp', 'w') as new_zip:
                    for file in zip_ref.infolist():
                        with zip_ref.open(file) as f:
                            if file.filename.endswith('.json'):
                                content = json.load(f)
                                new_content = json.dumps(content, indent=None).encode('utf-8')
                                new_zip.writestr(file, new_content)
                            else:
                                new_zip.writestr(file, f.read())
        except zipfile.BadZipFile:
            os.remove(file_path)
            print(f"Warning: '{file_path}' is not a zip file or is corrupted and will be removed.")
        os.replace(file_path + '.tmp', file_path)

def walk_and_process(root_dir):
    file_paths = []
    for root, dirs, files in os.walk(root_dir):
        for file in files:
            file_paths.append(os.path.join(root, file))

    with multiprocessing.Pool() as pool:
        list(tqdm(pool.imap(process_file, file_paths), total=len(file_paths)))

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Remove .lock files and rewrite .zip-contained JSONs without indentation.')
    parser.add_argument('root_dir', type=str, help='Root directory to start processing')
    
    args = parser.parse_args()
    
    walk_and_process(args.root_dir)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022, Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import json
from argparse import ArgumentParser

from oc_meta.plugins.analyser import OCMetaCounter

if __name__ == '__main__': # pragma: no cover
    arg_parser = ArgumentParser('meta_analyser.py', description='This script allows performing various analyses on the OC Meta output CSVs')
    arg_parser.add_argument('-c', '--csv_dir', dest='csv_dir', required=True, help='The directory containing the OC Meta output CSVs')
    arg_parser.add_argument('-w', '--what', dest='what', required=True, choices=['authors', 'editors', 'publishers', 'venues', 'types', 'years'], help='What to count')
    arg_parser.add_argument('-b', '--by', dest='by_what', required=False, default=None, choices=['publication', 'venue'], help="This parameter is only useful for the 'top' operation. It allows entities specified in 'what' to be ordered by other entities ('by what'). For example, sorting journals by number of publications")
    arg_parser.add_argument('-n', '--number', dest='number', required=False, default=None, type=int, help="This parameter is only useful for the 'top' operation. Once the entities are sorted, only the first N are returned. If N is not specified, all entities are returned")
    args = arg_parser.parse_args()
    analyser = OCMetaCounter(args.csv_dir)
    if args.what and not args.by_what:
        result = analyser.count(what=args.what)
    elif args.what and args.by_what:
        result = analyser.get_top(what=args.what, by_what=args.by_what, number=args.number)
    if isinstance(result, str):
        with open(f'{args.what}_count.txt', 'w', encoding='utf-8') as f:
            f.write(result)
    elif isinstance(result, list):
        with open(f'count_{args.what}_by_{args.by_what}.json', 'w', encoding='utf-8') as outfile:
            json.dump(result, outfile)


import argparse
import csv
from SPARQLWrapper import SPARQLWrapper, JSON
from tqdm import tqdm

def get_entity_type(entity_url):
    parts = entity_url.split('/')
    if 'oc' in parts and 'meta' in parts:
        try:
            return parts[parts.index('meta') + 1]
        except IndexError:
            return None
    return None

def query_sparql(endpoint_url, entity):
    sparql = SPARQLWrapper(endpoint_url)
    query = f"""
    PREFIX fabio: <http://purl.org/spar/fabio/>
    PREFIX frbr: <http://purl.org/vocab/frbr/core#>
    SELECT ?type ?partOfType
    WHERE {{
        <{entity}> a ?type .
        FILTER (?type != fabio:Expression)
        OPTIONAL {{
            <{entity}> frbr:partOf ?container .
            ?container a ?partOfType .
            FILTER (?partOfType != fabio:Expression)
        }}
    }}
    """
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    return sparql.query().convert()

def process_csv(input_csv, output_csv, endpoint_url):
    with open(input_csv, mode='r', newline='', encoding='utf-8') as infile, \
         open(output_csv, mode='w', newline='', encoding='utf-8') as outfile:

        reader = list(csv.DictReader(infile))
        writer = csv.DictWriter(outfile, fieldnames=['entity', 'type', 'partOfType'])
        writer.writeheader()

        for row in tqdm(reader, desc='Processing rows', unit='row'):
            entity = row['surviving_entity']
            if get_entity_type(entity) == 'br':
                results = query_sparql(endpoint_url, entity)
                for result in results["results"]["bindings"]:
                    writer.writerow({
                        'entity': entity,
                        'type': result["type"]["value"],
                        'partOfType': result["partOfType"]["value"] if "partOfType" in result else None
                    })

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Find the containment chain of bibliographic resources')
    parser.add_argument('input_csv', help='The path to the input CSV file')
    parser.add_argument('output_csv', help='The path to the output CSV file')
    parser.add_argument('endpoint_url', help='The SPARQL endpoint URL')
    args = parser.parse_args()

    process_csv(args.input_csv, args.output_csv, args.endpoint_url)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2019 Silvio Peroni <essepuntato@gmail.com>
# Copyright 2019-2020 Fabio Mariani <fabio.mariani555@gmail.com>
# Copyright 2021 Simone Persiani <iosonopersia@gmail.com>
# Copyright 2021-2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.


from __future__ import annotations

import os
import re
from argparse import ArgumentParser

from tqdm import tqdm

from oc_meta.lib.csvmanager import CSVManager
from oc_meta.lib.file_manager import get_csv_data
from oc_meta.lib.master_of_regex import name_and_ids


def map_metaid(ids: list, mapping: dict):
    ids = set(ids)
    metaid = [identifier for identifier in ids if identifier.split(':', maxsplit=1)[0] == 'omid'][0]
    mapping.setdefault(metaid, set())
    ids.remove(metaid)
    for other_id in ids:
        mapping[metaid].add(other_id)

def extract_metaid_mapping(input_dir:str, output_dir:str) -> None:
    mapping = dict()
    filenames = os.listdir(input_dir)
    pbar = tqdm(total=len(filenames))
    for filename in filenames:
        filepath = os.path.join(input_dir, filename)
        data = get_csv_data(filepath)
        for row in data:
            if row['id']:
                map_metaid(row['id'].split(), mapping)
                venue_name_and_ids = re.search(name_and_ids, row['venue'])
                if venue_name_and_ids:
                    map_metaid(venue_name_and_ids.group(2).split(), mapping)
                publisher_name_and_ids = re.search(name_and_ids, row['publisher'])
                if publisher_name_and_ids:
                    map_metaid(publisher_name_and_ids.group(2).split(), mapping)
                authors = row['author'].split('; ')
                editors = row['editor'].split('; ')
                for ra in [authors, editors]:
                    for individual in ra:
                        individual_name_and_ids = re.search(name_and_ids, individual)
                        if individual_name_and_ids:
                            map_metaid(individual_name_and_ids.group(2).split(), mapping)
        pbar.update()
    pbar.close()
    data_to_dump = CSVManager(output_path=output_dir)
    counter = 0
    threshold = 10000
    for metaid, ids in mapping.items():
        for identifier in ids:
            if counter > 0 and counter % threshold == 0:
                data_to_dump.dump_data(f'{counter - threshold + 1}-{counter}.csv')
            data_to_dump.add_value(metaid, identifier)
            counter += 1
    if data_to_dump.data:
        data_to_dump.dump_data(f'{counter - (counter % threshold) + 1}-{counter}.csv')


if __name__ == '__main__':
    arg_parser = ArgumentParser('Extract a mapping file between MetaIDs and other identifiers')
    arg_parser.add_argument('-i', '--input', dest='input', required=True,
                            help='The dirpath of the Meta output CSV files')
    arg_parser.add_argument('-o', '--output', dest='output', required=True,
                            help='The output dirpath')
    args = arg_parser.parse_args()
    extract_metaid_mapping(args.input, args.output)


import argparse
import json
from rdflib import URIRef
from oc_ocdm.support.support import find_paths
from time_agnostic_library.agnostic_query import VersionQuery
from tqdm import tqdm
from datetime import datetime
import multiprocessing

def process_manifestations(args):
    br_uri, rdf_dir, config_dict = args
    _, cur_file_path = find_paths(URIRef(f"{br_uri}/prov/se/1"), rdf_dir, "https://w3id.org/oc/meta/", "_", 10000, 1000, True, None)
    cur_file_path = cur_file_path.replace('.json', '.zip')
    config_dict['provenance']['file_paths'] = [cur_file_path]

    query = f'''
        SELECT DISTINCT ?re
        WHERE {{
            <{br_uri}> <http://purl.org/vocab/frbr/core#embodiment> ?re.
        }}
    '''
    agnostic_query = VersionQuery(query, None, False, config_path=None, config_dict=config_dict)
    agnostic_result, _ = agnostic_query.run_agnostic_query()
    sorted_agnostic_result = sorted(agnostic_result.items(), key=lambda x: datetime.strptime(x[0], "%Y-%m-%dT%H:%M:%S"))

    first_relevant_resources = None
    for _, resources in sorted_agnostic_result:
        if resources:
            first_relevant_resources = {res[0] for res in resources}
            break

    resources_to_delete = set()
    if first_relevant_resources:
        for _, resources in sorted_agnostic_result:
            if resources:
                current_resources = {res[0] for res in resources}
                resources_to_delete.update(current_resources - first_relevant_resources)

    return resources_to_delete

def main():
    parser = argparse.ArgumentParser(description="Remove multiple manifestations from bibliographic resources.")
    parser.add_argument("json_file", type=str, help="Path to the JSON file containing information about bibliographic resources with multiple manifestations")
    parser.add_argument("rdf_dir", type=str, help="Path to the OpenCitations Meta RDF directory")
    parser.add_argument("time_agnostic_config", type=str, help="Path to the time agnostic library configuration file")
    args = parser.parse_args()

    with open(args.json_file, 'r') as file:
        data = json.load(file)

    with open(args.time_agnostic_config, 'r') as f:
        config_dict = json.load(f)

    pool = multiprocessing.Pool(processes=multiprocessing.cpu_count())
    tasks = [(br_uri, args.rdf_dir, config_dict) for br_uri, _ in data.items()]
    results = list(tqdm(pool.imap_unordered(process_manifestations, tasks), total=len(tasks)))

    pool.close()
    pool.join()

    deletions = list()
    for result in results:
        deletions.extend(result)

    with open('re_to_be_deleted.json', 'w') as f:
        json.dump(deletions, f, indent=4)

    print(f"Results saved to 're_to_be_deleted.json'")


if __name__ == "__main__":
    main()


import argparse
import os
import multiprocessing
from zipfile import ZipFile
import rdflib
from tqdm import tqdm
import functools

def explore_dir(root_dir, entity_type):
    """Esplora ricorsivamente per trovare i file .zip validi, 
    escludendo se.zip e scegliendo la directory corretta in base al tipo di entità."""
    sub_dir = 'br' if entity_type == 'bibliographic_resources' else 'ar'
    target_path = os.path.join(root_dir, sub_dir)
    # Itera sui file nella directory target e nelle sue sottodirectory
    for root, _, files in os.walk(target_path):
        for filename in files:
            # Seleziona i file .zip escludendo 'se.zip'
            if filename.endswith('.zip') and filename != 'se.zip':
                yield os.path.join(root, filename)

def load_and_query_zip(zip_path, entity_type):
    """Carica e interroga il contenuto del file zip senza decomprimerlo."""
    g = rdflib.ConjunctiveGraph()
    with ZipFile(zip_path) as myzip:
        with myzip.open(myzip.namelist()[0]) as myfile:
            g.parse(myfile, format='json-ld')
    if entity_type == 'bibliographic_resources':
        query = """
        SELECT (COUNT(DISTINCT ?s) AS ?count) WHERE {
            ?s a <http://purl.org/spar/fabio/Expression> .
        }
        """
    elif entity_type == 'authors':
        query = """
        prefix pro: <http://purl.org/spar/pro/>
        SELECT (COUNT(DISTINCT ?s) AS ?count) WHERE {
            ?s a pro:RoleInTime;
            pro:withRole pro:author.
        }
        """
    elif entity_type == 'editors':
        query = """
        prefix pro: <http://purl.org/spar/pro/>
        SELECT (COUNT(DISTINCT ?s) AS ?count) WHERE {
            ?s a pro:RoleInTime;
            pro:withRole pro:editor.
        }
        """
    elif entity_type == 'publishers':
        query = """
        prefix pro: <http://purl.org/spar/pro/>
        SELECT (COUNT(DISTINCT ?s) AS ?count) WHERE {
            ?s a pro:RoleInTime;
            pro:withRole pro:publisher.
        }
        """
    for row in g.query(query):
        return int(row[0])

def main(root_dir, entity_type):
    zip_files = list(explore_dir(root_dir, entity_type))
    with multiprocessing.Pool(multiprocessing.cpu_count()) as pool:
        func = functools.partial(load_and_query_zip, entity_type=entity_type)
        results = list(tqdm(pool.imap(func, zip_files), total=len(zip_files)))
    total_count = sum(results)
    print(f"Totale entità di tipo Expression: {total_count}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Conta le risorse bibliografiche e gli autori, editori, o pubblicatori.')
    parser.add_argument('root_dir', type=str, help='La root directory da esplorare.')
    parser.add_argument('entity_type', type=str, choices=['bibliographic_resources', 'authors', 'editors', 'publishers'], 
                        help='Il tipo di entità da contare.')
    args = parser.parse_args()
    main(args.root_dir, args.entity_type)



import argparse
import os
import zipfile
from tqdm import tqdm
from multiprocessing import Pool, cpu_count

def extract_zip(args):
    file_path, input_path, output_path = args
    root = os.path.dirname(file_path)
    output_dir = root.replace(input_path, output_path, 1)
    try:
        os.makedirs(output_dir, exist_ok=True)  # exist_ok=True ignora l'errore se la cartella esiste già
    except FileExistsError:
        pass  # Se la cartella è stata creata da un altro processo nel frattempo, ignora l'errore
    with zipfile.ZipFile(file_path, 'r') as zip_ref:
        for member in zip_ref.namelist():
            extracted_path = os.path.join(output_dir, member)
            if not extracted_path.startswith(os.path.commonpath([output_dir])):
                raise Exception("Path traversal attempt detected")
            zip_ref.extract(member, output_dir)

def find_and_extract(input_path, output_path, data_only=False):
    args_list = []
    for root, dirs, files_in_dir in os.walk(input_path):
        for file in files_in_dir:
            if file.endswith(".zip"):
                file_path = os.path.join(root, file)
                path_parts = os.path.normpath(file_path).split(os.sep)
                if data_only and 'prov' not in path_parts:
                    args_list.append((file_path, input_path, output_path))
                elif not data_only:
                    args_list.append((file_path, input_path, output_path))
    
    # Utilizzo di multiprocessing per l'estrazione parallela
    pool = Pool(processes=cpu_count())
    for _ in tqdm(pool.imap_unordered(extract_zip, args_list), total=len(args_list), desc="Estrazione file in corso"):
        pass
    pool.close()
    pool.join()

def main():
    parser = argparse.ArgumentParser(description="Estrai file JSON da archivi ZIP in parallelo.")
    parser.add_argument("input_path", type=str, help="Percorso della cartella di input.")
    parser.add_argument("output_path", type=str, help="Percorso della cartella di output.")
    parser.add_argument("--data_only", action="store_true", help="Estrai file solo da sottocartelle diverse da 'prov'.")

    args = parser.parse_args()

    # Assicurazione che output_path termini con os.sep
    if not args.output_path.endswith(os.sep):
        args.output_path += os.sep

    find_and_extract(args.input_path, args.output_path, args.data_only)

if __name__ == "__main__":
    main()


import argparse
import gzip
import os
import zipfile
from pebble import ProcessPool
from concurrent.futures import TimeoutError
from multiprocessing import Manager
from tqdm import tqdm
import rdflib


def count_triples_in_file(compression_type, file_path, data_format):
    triple_count = 0
    try:
        if compression_type == 'zip':
            with zipfile.ZipFile(file_path, 'r') as z:
                jsonld_filename = [f for f in z.namelist()][0]
                with z.open(jsonld_filename) as f:
                    file_content = f.read().decode('utf-8')
        elif compression_type == 'gz':
            with gzip.open(file_path, 'rt') as f:
                file_content = f.read()
        elif compression_type in ['json', 'ttl']:
            with open(file_path, 'r', encoding='utf8') as f:
                file_content = f.read()
        g = rdflib.ConjunctiveGraph()
        g.parse(data=file_content, format=data_format)
        triple_count += len(g)
    except Exception as e:
        print(f"Error processing file in {file_path}: {e}")
    return triple_count

def process_directory(directory, compression_type, prov_only, data_only, data_format):
    if compression_type not in ['zip', 'gz', 'json', 'ttl']:
        raise ValueError("Unsupported compression type.")

    files = []
    for root, dirs, files_in_dir in os.walk(directory):
        for file in files_in_dir:
            if file.endswith('.' + compression_type):
                file_path = os.path.join(root, file)
                path_parts = os.path.normpath(file_path).split(os.sep)                
                if prov_only and 'prov' in path_parts:
                    files.append(file_path)
                elif data_only and 'prov' not in path_parts:
                    files.append(file_path)
                elif not prov_only and not data_only:
                    files.append(file_path)

    total_triples = 0
    with ProcessPool() as pool:
        future_results = []
        for file_path in files:
            future = pool.schedule(count_triples_in_file, args=(compression_type, file_path, data_format))
            future_results.append(future)

        for future in tqdm(future_results, desc="Processing files", unit="file"):
            try:
                result = future.result()
                total_triples += result
            except TimeoutError as error:
                print(f"Operation took longer than expected: {error}")
            except Exception as error:
                print(f"Error during execution: {error}")

    return total_triples

def main():
    parser = argparse.ArgumentParser(description="Conta le triple RDF nei file compressi in una directory.")
    parser.add_argument('directory', type=str, help='Directory da esplorare')
    parser.add_argument('compression_type', type=str, choices=['zip', 'gz', 'json', 'ttl'], help='Type of file compression (zip, gz, json, or nt)')
    parser.add_argument('data_format', type=str, choices=['json-ld', 'nquads', 'turtle'], help='Data format of the files (json-ld, nquads, or nt)')
    parser.add_argument('--prov_only', action='store_true', help='Conta solo nelle sottocartelle di nome prov', default=False)
    parser.add_argument('--data_only', action='store_true', help='Conta solo nelle sottocartelle di nome diverso da prov', default=False)

    args = parser.parse_args()

    total_triples = process_directory(args.directory, args.compression_type, args.prov_only, args.data_only, args.data_format)
    print(f"Numero totale di triple RDF: {total_triples}")

if __name__ == "__main__":
    main()


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE


import os
import shutil
from argparse import ArgumentParser

import yaml

from oc_meta.lib.file_manager import normalize_path
from oc_meta.plugins.multiprocess.prepare_multiprocess import (
    prepare_relevant_items, split_csvs_in_chunks)
from oc_meta.run.meta_process import MetaProcess, run_meta_process

if __name__ == '__main__': # pragma: no cover
    arg_parser = ArgumentParser('prepare_multiprocess.py', description='Venues, authors and editors are preprocessed not to create duplicates when running Meta in multi-process')
    arg_parser.add_argument('-c', '--config', dest='config', required=True, help='Configuration file path')
    arg_parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', required=False, help='Show a loading bar, elapsed and estimated time')
    args = arg_parser.parse_args()
    config = args.config
    with open(config, encoding='utf-8') as file:
        settings = yaml.full_load(file)
    csv_dir = normalize_path(settings['input_csv_dir'])
    items_per_file = settings['items_per_file']
    workers_numbers = settings['workers_number']
    verbose = args.verbose
    meta_process = MetaProcess(settings=settings, meta_config_path=config)
    TMP_DIR = os.path.join(meta_process.base_output_dir, 'tmp')
    csv_dir = csv_dir.rstrip('/')
    if not os.path.isdir(meta_process.input_csv_dir + '_old'):
        split_csvs_in_chunks(csv_dir=csv_dir, output_dir=TMP_DIR, chunk_size=1000, verbose=verbose)
        os.rename(csv_dir, csv_dir + '_old')
        os.mkdir(csv_dir)
        for file in os.listdir(TMP_DIR):
            shutil.move(os.path.join(TMP_DIR, file), csv_dir)
    if not any(os.path.exists(os.path.join(TMP_DIR, directory)) for directory in ['venues', 'ids', 'publishers', 'authors', 'editors']):
        prepare_relevant_items(csv_dir=csv_dir, output_dir=TMP_DIR, items_per_file=items_per_file, verbose=verbose)
    for resp_agent in ['authors', 'editors', 'publishers']:
        resp_agent_dir = os.path.join(TMP_DIR, resp_agent)
        if os.path.isdir(resp_agent_dir):
            settings['input_csv_dir'] = resp_agent_dir
            run_meta_process(settings=settings, meta_config_path=config, resp_agents_only=True)
    settings['workers_number'] = 1
    for entity_type in ['venues', 'ids']:
        entity_dir = os.path.join(TMP_DIR, entity_type)
        if os.path.isdir(entity_dir):
            settings['input_csv_dir'] = entity_dir
            run_meta_process(settings=settings, meta_config_path=config)


import argparse
import zipfile
from tqdm import tqdm
from multiprocessing import Pool
import rdflib
from pathlib import Path
import os


def extract_author_uris(zip_path):
    graph = rdflib.ConjunctiveGraph()
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        with zip_ref.open(zip_ref.namelist()[0]) as rdf_file:
            graph.parse(rdf_file, format='json-ld')
                        
    query = """
    SELECT DISTINCT ?ra WHERE {
        ?ar <http://purl.org/spar/pro/withRole> <http://purl.org/spar/pro/author> ;
           <http://purl.org/spar/pro/isHeldBy> ?ra .
    }
    """
    results = graph.query(query)
    return {str(row.ra) for row in results}

def extract_identified_ras(zip_path):
    graph = rdflib.ConjunctiveGraph()
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        with zip_ref.open(zip_ref.namelist()[0]) as rdf_file:
            graph.parse(rdf_file, format='json-ld')

    query = """
    SELECT DISTINCT ?ra WHERE {
        ?ra <http://purl.org/spar/datacite/hasIdentifier> ?id .
    }
    """
    results = graph.query(query)
    return {str(row.ra) for row in results}

def process_directory(directory):
    path_ar = Path(os.path.join(directory, 'ar'))
    zip_files_ar = [f for f in path_ar.rglob('*.zip') if f.name != 'se.zip']
    
    with Pool() as pool:
        authors = set.union(*list(tqdm(pool.imap(extract_author_uris, zip_files_ar), total=len(zip_files_ar), desc="Looking for authors")))
    
    path_ra = Path(os.path.join(directory, 'ra'))
    zip_files_ra = [f for f in path_ra.rglob('*.zip') if f.name != 'se.zip']

    with Pool() as pool:
        identified_ras = set.union(*list(tqdm(pool.imap(extract_identified_ras, zip_files_ra), total=len(zip_files_ra), desc="Looking for identified responsible agents")))

    total_authors = len(authors)
    identified_authors = len(authors.intersection(identified_ras))
    percentage = (identified_authors / total_authors * 100) if total_authors > 0 else 0
    
    return total_authors, identified_authors, percentage

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Analyze RDF data within ZIP files to identify and count authors with specific roles and identifiers.")
    parser.add_argument("directory", help="Path to the directory containing ZIP files for analysis")
    args = parser.parse_args()
    
    total_authors, identified_authors, percentage = process_directory(args.directory)
    print(f"There are {total_authors} authors, of which {identified_authors} have an identifier ({percentage:.2f}%).")


import argparse
import os
import zipfile
import json
import redis
from SPARQLWrapper import SPARQLWrapper, JSON

def create_redis_connection():
    return redis.Redis(host='localhost', port=6379, db=14)

def extract_omid(entity):
    parts = entity.split('/')
    return f"omid:{parts[-2]}/{parts[-1]}", parts[-2]

def search_omid_in_redis(r, omid):
    result = r.get(omid)
    return result.decode() if result else None

def run_sparql_query(sparql_endpoint, entity_id, entity_class):
    sparql = SPARQLWrapper(sparql_endpoint)
    if entity_class == "id":
        query = f"""
            PREFIX datacite: <http://purl.org/spar/datacite/>
            SELECT ?br WHERE {{
                <{entity_id}> ^datacite:hasIdentifier ?br.
            }}
        """
    elif entity_class == "ar":
        query = f"""
            PREFIX pro: <http://purl.org/spar/pro/>
            SELECT ?br WHERE {{
                <{entity_id}> ^pro:isDocumentContextFor ?br.
            }}
        """
    elif entity_class == "ra":
        query = f"""
            PREFIX pro: <http://purl.org/spar/pro/>
            SELECT ?br WHERE {{
                <{entity_id}> ^pro:isHeldBy/isDocumentContextFor ?br.
            }}
        """
    elif entity_class == "re":
        query = f"""
            PREFIX frbr: <http://purl.org/vocab/frbr/core#>
            SELECT ?br WHERE {{
                <{entity_id}> ^frbr:embodiment ?br.
            }}
        """
    elif entity_class == "br":
        query = f"""
            SELECT ?type WHERE {{
                <{entity_id}> a ?type.
            }}
        """
        sparql.setQuery(query)
        sparql.setReturnFormat(JSON)
        results = sparql.query().convert()
        for result in results["results"]["bindings"]:
            br_type = result["type"]["value"]
        if br_type in {'http://purl.org/spar/fabio/JournalIssue', 'http://purl.org/spar/fabio/JournalVolume', 'http://purl.org/spar/fabio/Journal'}:
            query = f"""
                PREFIX fabio: <http://purl.org/spar/fabio/>
                PREFIX frbr: <http://purl.org/vocab/frbr/core#>
                SELECT ?br WHERE {{
                    <{entity_id}> ^frbr:partOf* ?br.
                    ?br a fabio:JournalArticle.
                }}
            """
            sparql.setQuery(query)
            sparql.setReturnFormat(JSON)
            results = sparql.query().convert()
            for result in results["results"]["bindings"]:
                return result["br"]["value"]
        else:
            query = f"""
                PREFIX fabio: <http://purl.org/spar/fabio/>
                PREFIX frbr: <http://purl.org/vocab/frbr/core#>
                SELECT ?br WHERE {{
                    <{entity_id}> ^frbr:partOf ?br.
                }}
            """
            sparql.setQuery(query)
            sparql.setReturnFormat(JSON)
            results = sparql.query().convert()
            for result in results["results"]["bindings"]:
                return result["br"]["value"]
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    results = sparql.query().convert()
    for result in results["results"]["bindings"]:
        return result["br"]["value"]

def explore_folder(folder_path, sparql_endpoint, r):
    for root, dirs, files in os.walk(folder_path):
        for file in files:
            if file == 'se.zip':
                file_path = os.path.join(root, file)
                print(file_path)
                # with zipfile.ZipFile(file_path, 'r') as zip_ref:
                #     for zip_info in zip_ref.infolist():
                #         with zip_ref.open(zip_info) as f:
                #             data = json.loads(f.read().decode())
                #             for graph in data:
                #                 for entity in graph['@graph']:
                #                     entity_id = entity['http://www.w3.org/ns/prov#specializationOf'][0]['@id']
                #                     if 'http://www.w3.org/ns/prov#hadPrimarySource' not in entity:
                #                         omid, entity_class = extract_omid(entity_id)
                #                         primary_source = search_omid_in_redis(r, omid)
                #                         if primary_source is None:
                #                             sparql_result = run_sparql_query(sparql_endpoint, entity_id, entity_class)
                #                             if sparql_result:
                #                                 br, _ = extract_omid(sparql_result)
                #                                 primary_source = search_omid_in_redis(r, br)
                                            

def main():
    parser = argparse.ArgumentParser(description="Explore folder for se.zip files and display their contents.")
    parser.add_argument("folder_path", type=str, help="Path of the folder to explore")
    parser.add_argument("sparql_endpoint", type=str, help="SPARQL endpoint URL")
    args = parser.parse_args()
    r = create_redis_connection()
    explore_folder(args.folder_path, args.sparql_endpoint, r)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import json
import os
import zipfile
from collections import defaultdict
from tqdm import tqdm

import rdflib


def merge_file_set(identifier, file_paths, zip_output_rdf):
    """
    Merges partial RDF files into a single final file with progress tracking
    """
    try:
        merged_graph = rdflib.ConjunctiveGraph()
        base_file_path = f"{identifier}.json" if not zip_output_rdf else f"{identifier}.zip"

        if os.path.exists(base_file_path):
            if zip_output_rdf:
                with zipfile.ZipFile(base_file_path, 'r') as zipf:
                    with zipf.open(zipf.namelist()[0], 'r') as json_file:
                        base_graph_data = json.load(json_file)
                        merged_graph.parse(data=json.dumps(base_graph_data), format='json-ld')
            else:
                with open(base_file_path, 'r', encoding='utf-8') as json_file:
                    base_graph_data = json.load(json_file)
                    merged_graph.parse(data=json.dumps(base_graph_data), format='json-ld')
        for file_path in file_paths:
            if zip_output_rdf:
                with zipfile.ZipFile(file_path, 'r') as zipf:
                    with zipf.open(zipf.namelist()[0], 'r') as json_file:
                        graph_data = json.load(json_file)
                        merged_graph.parse(data=json.dumps(graph_data), format='json-ld')
            else:
                with open(file_path, 'r', encoding='utf-8') as json_file:
                    graph_data = json.load(json_file)
                    merged_graph.parse(data=json.dumps(graph_data), format='json-ld')

        if zip_output_rdf:
            with zipfile.ZipFile(base_file_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
                zipf.writestr(f"{os.path.basename(identifier)}.json", 
                            merged_graph.serialize(format='json-ld', indent=None, ensure_ascii=False).encode('utf-8'))
        else:
            with open(base_file_path, 'w', encoding='utf-8') as f:
                f.write(merged_graph.serialize(format='json-ld', indent=None, ensure_ascii=False))

        for file_path in file_paths:
            os.remove(file_path)

        print(f"Merge completed for: {base_file_path}")
    except Exception as e:
        print(f"Error during file merge for {identifier}: {e}")


def find_unmerged_files(base_dir):
    """
    Recursively explores base_dir to find unmerged partial RDF files
    """
    files_to_merge = defaultdict(list)

    for root, dirs, files in os.walk(base_dir):
        json_files = [f for f in files if f.endswith('.json')]
        zip_files = [f for f in files if f.endswith('.zip')]

        for file in json_files:
            if '_' in file:
                number_part = file.split('_')[0]
                identifier = os.path.join(root, number_part)
                files_to_merge[identifier].append(os.path.join(root, file))

        for file in zip_files:
            if '_' in file:
                number_part = file.split('_')[0]
                identifier = os.path.join(root, number_part)
                files_to_merge[identifier].append(os.path.join(root, file))

    final_candidates = {}
    for identifier, partial_files in files_to_merge.items():
        zip_output_rdf = any(pf.endswith('.zip') for pf in partial_files)
        final_file = f"{identifier}.{'zip' if zip_output_rdf else 'json'}"
        
        if not os.path.exists(final_file) or partial_files:
            final_candidates[identifier] = (partial_files, zip_output_rdf)

    return final_candidates


def main():
    parser = argparse.ArgumentParser(description='Finds and merges unmerged RDF files.')
    parser.add_argument('base_dir', help='Base directory to search for unmerged files.')
    args = parser.parse_args()

    anomalies = find_unmerged_files(args.base_dir)
    if not anomalies:
        print("No anomalies found.")
    else:
        print(f"Found {len(anomalies)} sets of files to merge:")
        for identifier, (partial_files, zip_output_rdf) in tqdm(anomalies.items(), desc="Processing file sets", unit="set"):
            merge_file_set(identifier, partial_files, zip_output_rdf)


if __name__ == "__main__":
    main()


import argparse
import gzip
from pathlib import Path
import rdflib
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor
import multiprocessing

def load_and_search(file_path, entity_uri):
    """
    Load a .jsonld.gz file into an RDF graph and search for the entity URI.

    Parameters:
    - file_path: The path to the .jsonld.gz file.
    - entity_uri: The URI of the entity to search for.

    Returns:
    - The file path if the entity is found, otherwise None.
    """
    graph = rdflib.ConjunctiveGraph()
    with gzip.open(file_path, 'rt', encoding='utf-8') as file:
        graph.parse(file, format='json-ld')

    query = """
    ASK {
        <""" + entity_uri + """> ?p ?o .
    }
    """
    if graph.query(query).askAnswer:
        return str(file_path)
    return None

def main(directory, entity_uri):
    """
    Search for an entity URI in .jsonld.gz files within the specified directory using multiprocessing.

    Parameters:
    - directory: The path to the directory containing .jsonld.gz files.
    - entity_uri: The URI of the entity to search for.
    """
    path = Path(directory)
    files = [str(file_path) for file_path in path.glob('*.jsonld.gz')]
    
    # Determine the number of processes to use
    num_processes = min(multiprocessing.cpu_count(), len(files))

    found_files = []
    with ProcessPoolExecutor(max_workers=num_processes) as executor:
        results = list(tqdm(executor.map(load_and_search, files, [entity_uri]*len(files)), total=len(files), desc="Processing files"))

    # Filter out None results
    found_files = [result for result in results if result]

    if found_files:
        print("Entity found in the following files:")
        for file in found_files:
            print(file)
    else:
        print("Entity not found in any file.")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Search for an entity URI in .jsonld.gz files with multiprocessing.")
    parser.add_argument("directory", type=str, help="The path to the directory containing .jsonld.gz files.")
    parser.add_argument("entity_uri", type=str, help="The URI of the entity to search for.")
    args = parser.parse_args()

    main(args.directory, args.entity_uri)


import redis
from tqdm import tqdm
import multiprocessing

redis_client = redis.StrictRedis(host='localhost', port=6379, db=14)

keys = redis_client.keys('*')

count = 0

progress_bar = tqdm(total=len(keys), desc="Conteggio chiavi")

for key in keys:
    value = redis_client.get(key)
    if value.decode('utf-8') == 'https://api.japanlinkcenter.org/':
        count += 1
    progress_bar.update(1)
progress_bar.close()

print(f"Numero di chiavi con il valore 'https://api.japanlinkcenter.org/': {count}")


import os
import sys


def trova_file_mancanti(file_txt):
    elenco_file = []
    file_mancanti = []

    # Legge il file txt e crea l'elenco dei file
    with open(file_txt, "r") as file:
        elenco_file = file.read().splitlines()

    # Estrae i numeri dai nomi dei file e li converte in interi
    numeri_file = [int(os.path.splitext(nome_file)[0]) for nome_file in elenco_file]

    # Trova i file mancanti nella sequenza numerica
    ultimo_numero = max(numeri_file)
    for numero in range(1, ultimo_numero + 1):
        if numero not in numeri_file:
            file_mancanti.append(str(numero) + ".csv")

    return file_mancanti

# Verifica se è stato specificato il nome del file txt come argomento
if len(sys.argv) < 2:
    print("Usage: python script.py file.txt")
    sys.exit(1)

# Ottiene il nome del file txt dalla riga di comando
file_txt = sys.argv[1]

# Trova i file mancanti nella sequenza numerica
file_mancanti = trova_file_mancanti(file_txt)

# Stampa l'elenco dei file mancanti
if file_mancanti:
    print("File mancanti:")
    for file in file_mancanti:
        print(file)
else:
    print("Nessun file mancante nella sequenza numerica.")


import os
import json
import redis
from tqdm import tqdm
from multiprocessing import Pool, cpu_count

def process_file(args):
    file_path, redis_host, redis_port, db_number, omid_db_number = args
    r = redis.Redis(host=redis_host, port=redis_port, db=db_number)
    r_omid = redis.Redis(host=redis_host, port=redis_port, db=omid_db_number)
    doi_data = {}

    with open(file_path, 'r', encoding='utf8') as file:
        data = json.load(file)
        for item in data['items']:
            doi = str(item['DOI'])
            omid = r_omid.get(f'doi:{doi.lower()}')
            if omid:
                doi_data[f'omid:{omid.decode()}'] = 'https://api.crossref.org/snapshots/monthly/2023/09/all.json.tar.gz'
                if len(doi_data) >= 1000:
                    r.mset(doi_data)
                    doi_data = {}

    if doi_data:
        r.mset(doi_data)

def load_json_to_redis(folder_path, redis_host='localhost', redis_port=6379, db_number=14, omid_db_number=10, workers=cpu_count()):
    json_files = [os.path.join(folder_path, f) for f in os.listdir(folder_path) if f.endswith('.json')]
    num_processes = min(len(json_files), workers)
    with Pool(processes=num_processes) as pool:
        tasks = [(file, redis_host, redis_port, db_number, omid_db_number) for file in json_files]
        list(tqdm(pool.imap_unordered(process_file, tasks), total=len(tasks), desc="Caricamento", unit="file"))

    print("Caricamento completato.")

folder_path = '/srv/data/arcangelo/crossref-data-2023-09'
load_json_to_redis(folder_path, workers=64)


import argparse
import csv
import os
import random
import re
import time
import zipfile
from functools import partial
from multiprocessing import Pool, cpu_count

import filelock
import yaml
from oc_meta.plugins.editor import MetaEditor
from rdflib import RDF, ConjunctiveGraph, Literal, Namespace, URIRef
from SPARQLWrapper import JSON, SPARQLWrapper
from tqdm import tqdm

DATACITE = "http://purl.org/spar/datacite/"
FOAF = Namespace("http://xmlns.com/foaf/0.1/")
PROV = Namespace("http://www.w3.org/ns/prov#")
DCTERMS = Namespace("http://purl.org/dc/terms/")

def read_csv(csv_file):
    with open(csv_file, 'r') as f:
        reader = csv.DictReader(f)
        return list(reader)

def sparql_query_with_retry(sparql, max_retries=3, initial_delay=1, backoff_factor=2):
    for attempt in range(max_retries):
        try:
            return sparql.query().convert()
        except Exception as e:
            if attempt == max_retries - 1:
                raise
            delay = initial_delay * (backoff_factor ** attempt)
            time.sleep(delay + random.uniform(0, 1))

def check_agent_constraints(g: ConjunctiveGraph, entity):
    issues = []
    
    # Check type (must be exactly one: foaf:Agent)
    types = list(g.objects(entity, RDF.type, unique=True))
    if not types:
        issues.append(f"Entity {entity} has no type")
    elif len(types) > 1:
        issues.append(f"Entity {entity} has multiple types")
    elif URIRef(FOAF + "Agent") not in types:
        issues.append(f"Entity {entity} is not a foaf:Agent")
    
    # Check identifiers
    identifiers = list(g.objects(entity, URIRef(DATACITE + "hasIdentifier"), unique=True))
    if not identifiers:
        issues.append(f"Entity {entity} has no datacite:hasIdentifier")
    
    # Check name properties (must have at least one)
    has_name = False
    names = list(g.objects(entity, FOAF.name, unique=True))
    given_names = list(g.objects(entity, FOAF.givenName, unique=True))
    family_names = list(g.objects(entity, FOAF.familyName, unique=True))
    
    if names or given_names or family_names:
        has_name = True
    
    if not has_name:
        issues.append(f"Entity {entity} has no name properties (name, givenName, or familyName)")
        
    return issues

def check_entity_sparql(sparql_endpoint, entity_uri, is_surviving):
    sparql = SPARQLWrapper(sparql_endpoint)
    has_issues = False
    
    # Query to check if the entity exists
    exists_query = f"""
    ASK {{
        <{entity_uri}> ?p ?o .
    }}
    """
    sparql.setQuery(exists_query)
    sparql.setReturnFormat(JSON)
    exists_results = sparql_query_with_retry(sparql)

    if exists_results['boolean']:
        if not is_surviving:
            tqdm.write(f"Error in SPARQL: Merged entity {entity_uri} still exists")
            has_issues = True
    else:
        if is_surviving:
            tqdm.write(f"Error in SPARQL: Surviving entity {entity_uri} does not exist")
            has_issues = True
        return has_issues

    # Check type constraints
    types_query = f"""
    SELECT ?type WHERE {{
        <{entity_uri}> a ?type .
    }}
    """
    sparql.setQuery(types_query)
    sparql.setReturnFormat(JSON)
    types_results = sparql_query_with_retry(sparql)

    types = [result['type']['value'] for result in types_results['results']['bindings']]
    if not types:
        tqdm.write(f"Error in SPARQL: Entity {entity_uri} has no type")
        has_issues = True
    elif len(types) > 1:
        tqdm.write(f"Error in SPARQL: Entity {entity_uri} has multiple types")
        has_issues = True
    elif FOAF + "Agent" not in types:
        tqdm.write(f"Error in SPARQL: Entity {entity_uri} is not a foaf:Agent")
        has_issues = True

    # Check name properties
    names_query = f"""
    SELECT ?name ?givenName ?familyName WHERE {{
        OPTIONAL {{ <{entity_uri}> <{FOAF}name> ?name }}
        OPTIONAL {{ <{entity_uri}> <{FOAF}givenName> ?givenName }}
        OPTIONAL {{ <{entity_uri}> <{FOAF}familyName> ?familyName }}
    }}
    """
    sparql.setQuery(names_query)
    sparql.setReturnFormat(JSON)
    names_results = sparql_query_with_retry(sparql)

    has_name = False
    for result in names_results['results']['bindings']:
        if any(key in result for key in ['name', 'givenName', 'familyName']):
            has_name = True
            break

    if not has_name:
        tqdm.write(f"Error in SPARQL: Entity {entity_uri} has no name properties")
        has_issues = True

    # Check identifiers
    identifiers_query = f"""
    SELECT ?identifier WHERE {{
        <{entity_uri}> <{DATACITE}hasIdentifier> ?identifier .
    }}
    """
    sparql.setQuery(identifiers_query)
    sparql.setReturnFormat(JSON)
    identifiers_results = sparql_query_with_retry(sparql)

    identifiers = [result['identifier']['value'] for result in identifiers_results['results']['bindings']]
    if not identifiers:
        tqdm.write(f"Error in SPARQL: Entity {entity_uri} has no datacite:hasIdentifier")
        has_issues = True

    return has_issues

def get_entity_triples(sparql_endpoint, entity_uri):
    sparql = SPARQLWrapper(sparql_endpoint)
    query = f"""
    SELECT ?g ?s ?p ?o
    WHERE {{
        GRAPH ?g {{
            {{
                <{entity_uri}> ?p ?o .
                BIND(<{entity_uri}> AS ?s)
            }}
            UNION
            {{
                ?s ?p <{entity_uri}> .
                BIND(<{entity_uri}> AS ?o)
            }}
        }}
    }}
    """
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    results = sparql_query_with_retry(sparql)
    
    triples = []
    for result in results["results"]["bindings"]:
        graph = result["g"]["value"]
        subject = URIRef(result["s"]["value"])
        predicate = URIRef(result["p"]["value"])
        
        obj_data = result["o"]
        if obj_data["type"] == "uri":
            obj = URIRef(obj_data["value"])
        else:
            datatype = obj_data.get("datatype")
            obj = Literal(obj_data["value"], datatype=URIRef(datatype) if datatype else None)
        
        triples.append((graph, subject, predicate, obj))
    
    return triples

def check_entity_provenance(entity_uri, is_surviving, prov_graph: ConjunctiveGraph, prov_file_path):
    def extract_snapshot_number(snapshot_uri):
        match = re.search(r'/prov/se/(\d+)$', str(snapshot_uri))
        if match:
            return int(match.group(1))
        return 0

    snapshots = list(prov_graph.subjects(PROV.specializationOf, entity_uri, unique=True))
    if len(snapshots) <= 1:
        tqdm.write(f"Error in provenance file {prov_file_path}: Less than two snapshots found for entity {entity_uri}")
        return
    
    snapshots.sort(key=extract_snapshot_number)
    for i, snapshot in enumerate(snapshots):
        snapshot_number = extract_snapshot_number(snapshot)
        if snapshot_number != i + 1:
            tqdm.write(f"Error in provenance file {prov_file_path}: Snapshot {snapshot} has unexpected number {snapshot_number}, expected {i + 1}")
        
        gen_time = prov_graph.value(snapshot, PROV.generatedAtTime)
        if gen_time is None:
            tqdm.write(f"Error in provenance file {prov_file_path}: Snapshot {snapshot} has no generation time")
        
        if i < len(snapshots) - 1 or not is_surviving:
            invalidation_time = prov_graph.value(snapshot, PROV.invalidatedAtTime)
            if invalidation_time is None:
                tqdm.write(f"Error in provenance file {prov_file_path}: Non-last snapshot {snapshot} has no invalidation time")
        elif is_surviving and prov_graph.value(snapshot, PROV.invalidatedAtTime) is not None:
            tqdm.write(f"Error in provenance file {prov_file_path}: Last snapshot of surviving entity {snapshot} should not have invalidation time")

        description = prov_graph.value(snapshot, DCTERMS.description)
        is_merge_snapshot = description and "has been merged with" in str(description)
        
        derived_from = list(prov_graph.objects(snapshot, PROV.wasDerivedFrom, unique=True))
        if i == 0:  # First snapshot
            if derived_from:
                tqdm.write(f"Error in provenance file {prov_file_path}: First snapshot {snapshot} should not have prov:wasDerivedFrom relation")
        elif is_merge_snapshot:
            if len(derived_from) < 2:
                tqdm.write(f"Error in provenance file {prov_file_path}: Merge snapshot {snapshot} should be derived from at least two snapshots")
        else:  # Regular modification snapshot
            if len(derived_from) != 1:
                tqdm.write(f"Error in provenance file {prov_file_path}: Regular modification snapshot {snapshot} should have exactly one prov:wasDerivedFrom relation")
            else:
                previous_snapshot = snapshots[i-1]
                if derived_from[0] != previous_snapshot:
                    tqdm.write(f"Error in provenance file {prov_file_path}: Snapshot {snapshot} is not derived from the previous snapshot")

def process_file_group(args):
    file_path, entities, sparql_endpoint, query_output_dir = args

    if file_path is None:
        for entity, is_surviving, surviving_entity in entities:
            tqdm.write(f"Error: Could not find file for entity {entity}")
            has_issues = check_entity_sparql(sparql_endpoint, entity, is_surviving)
            if has_issues and not is_surviving:
                triples = get_entity_triples(sparql_endpoint, entity)
                combined_query = generate_update_query(entity, surviving_entity, triples)
                query_file_path = os.path.join(query_output_dir, f"update_{entity.split('/')[-1]}.sparql")
                with open(query_file_path, 'w') as f:
                    f.write(combined_query)
        return

    # Create lock files
    data_lock_file = f"{file_path}.lock"
    prov_lock_file = f"{file_path.replace('.zip', '')}/prov/se.zip.lock"
    
    data_lock = filelock.FileLock(data_lock_file)
    prov_lock = filelock.FileLock(prov_lock_file)

    try:
        with data_lock.acquire(timeout=60):  # Wait up to 60 seconds for the lock
            try:
                with zipfile.ZipFile(file_path, 'r') as zip_ref:
                    g = ConjunctiveGraph()
                    for filename in zip_ref.namelist():
                        with zip_ref.open(filename) as file:
                            g.parse(file, format='json-ld')
            except FileNotFoundError:
                for entity, is_surviving, surviving_entity in entities:
                    tqdm.write(f"Error: File not found for entity {entity}")
                    has_issues = check_entity_sparql(sparql_endpoint, entity, is_surviving)
                    if has_issues and not is_surviving:
                        triples = get_entity_triples(sparql_endpoint, entity)
                        combined_query = generate_update_query(entity, surviving_entity, triples)
                        query_file_path = os.path.join(query_output_dir, f"update_{entity.split('/')[-1]}.sparql")
                        with open(query_file_path, 'w') as f:
                            f.write(combined_query)
                return

            prov_file_path = file_path.replace('.zip', '') + '/prov/se.zip'
            try:
                with prov_lock.acquire(timeout=60):  # Wait up to 60 seconds for the lock
                    try:
                        with zipfile.ZipFile(prov_file_path, 'r') as zip_ref:
                            prov_graph = ConjunctiveGraph()
                            for filename in zip_ref.namelist():
                                with zip_ref.open(filename) as file:
                                    prov_graph.parse(file, format='json-ld')
                    except FileNotFoundError:
                        prov_graph = None
                    except zipfile.BadZipFile:
                        prov_graph = "badzip"

                    for entity, is_surviving, surviving_entity in entities:
                        if (URIRef(entity), None, None) not in g:
                            if is_surviving:
                                tqdm.write(f"Error in file {file_path}: Surviving entity {entity} does not exist")
                        else:
                            if not is_surviving:
                                tqdm.write(f"Error in file {file_path}: Merged entity {entity} still exists")
                            else:
                                agent_issues = check_agent_constraints(g, URIRef(entity))
                                for issue in agent_issues:
                                    tqdm.write(f"Error in file {file_path}: {issue}")

                        if prov_graph == None:
                            tqdm.write(f"Error: Provenance file not found for entity {entity}")
                        elif prov_graph == "badzip":
                            tqdm.write(f"Error: Invalid zip file for provenance of entity {entity}")
                        else:
                            check_entity_provenance(URIRef(entity), is_surviving, prov_graph, prov_file_path)

                        # has_issues = check_entity_sparql(sparql_endpoint, entity, is_surviving)
                        # if has_issues and not is_surviving:
                        #     triples = get_entity_triples(sparql_endpoint, entity)
                        #     combined_query = generate_update_query(entity, surviving_entity, triples)
                        #     query_file_path = os.path.join(query_output_dir, f"update_{entity.split('/')[-1]}.sparql")
                        #     with open(query_file_path, 'w') as f:
                        #         f.write(combined_query)

            except filelock.Timeout:
                tqdm.write(f"Could not acquire lock for provenance file {prov_file_path} within timeout period")

    except filelock.Timeout:
        tqdm.write(f"Could not acquire lock for data file {file_path} within timeout period")

def generate_update_query(merged_entity, surviving_entity, triples):
    delete_query = "DELETE DATA {\n"
    insert_query = "INSERT DATA {\n"
    
    for graph, subject, predicate, obj in triples:
        if subject == URIRef(merged_entity):
            delete_query += f"  GRAPH <{graph}> {{ <{subject}> <{predicate}> {obj.n3()} }}\n"
        elif obj == URIRef(merged_entity):
            delete_query += f"  GRAPH <{graph}> {{ <{subject}> <{predicate}> <{obj}> }}\n"
            insert_query += f"  GRAPH <{graph}> {{ <{subject}> <{predicate}> <{surviving_entity}> }}\n"
    
    delete_query += "}\n"
    insert_query += "}\n"
    
    return delete_query + "\n" + insert_query

def process_csv(args, csv_file):
    csv_path, rdf_dir, meta_config_path, sparql_endpoint, query_output_dir = args
    csv_path = os.path.join(csv_path, csv_file)
    data = read_csv(csv_path)
    tasks = []

    meta_editor = MetaEditor(meta_config_path, "")

    for row in data:
        if 'Done' not in row or row['Done'] != 'True':
            continue

        surviving_entity = row['surviving_entity']
        merged_entities = row['merged_entities'].split('; ')
        all_entities = [surviving_entity] + merged_entities

        for entity in all_entities:
            file_path = meta_editor.find_file(rdf_dir, meta_editor.dir_split, meta_editor.n_file_item, entity, True)
            tasks.append((entity, entity == surviving_entity, file_path, rdf_dir, meta_config_path, sparql_endpoint, query_output_dir, surviving_entity))
    return tasks

def main():
    parser = argparse.ArgumentParser(description="Check merge process success on files and SPARQL endpoint for responsible agents")
    parser.add_argument('csv_folder', type=str, help="Path to the folder containing CSV files")
    parser.add_argument('rdf_dir', type=str, help="Path to the RDF directory")
    parser.add_argument('--meta_config', type=str, required=True, help="Path to meta configuration file")
    parser.add_argument('--query_output', type=str, required=True, help="Path to the folder where SPARQL queries will be saved")
    args = parser.parse_args()

    with open(args.meta_config, 'r') as config_file:
        config = yaml.safe_load(config_file)

    sparql_endpoint = config['triplestore_url']
    os.makedirs(args.query_output, exist_ok=True)

    csv_files = [f for f in os.listdir(args.csv_folder) if f.endswith('.csv')]
    
    # Process CSV files to gather tasks
    with Pool(processes=cpu_count()) as pool:
        process_csv_partial = partial(process_csv, (args.csv_folder, args.rdf_dir, args.meta_config, sparql_endpoint, args.query_output))
        all_tasks_list = list(tqdm(pool.imap(process_csv_partial, csv_files), total=len(csv_files), desc="Processing CSV files"))
    
    # Flatten the list of tasks
    all_tasks = [task for sublist in all_tasks_list for task in sublist]

    # Group tasks by file path
    tasks_by_file = {}
    for entity, is_surviving, file_path, rdf_dir, meta_config_path, sparql_endpoint, query_output_dir, surviving_entity in all_tasks:
        if file_path not in tasks_by_file:
            tasks_by_file[file_path] = []
        tasks_by_file[file_path].append((entity, is_surviving, surviving_entity))

    # Convert to list of arguments for parallel processing
    file_groups = [(file_path, tasks, sparql_endpoint, args.query_output) for file_path, tasks in tasks_by_file.items()]

    # Process each file group in parallel
    with Pool(processes=cpu_count()) as pool:
        list(tqdm(pool.imap(process_file_group, file_groups), total=len(file_groups), desc="Processing files"))

if __name__ == "__main__":
    main()


import csv
import os
from collections import Counter
from concurrent.futures import ProcessPoolExecutor

import pycountry
from langdetect import detect
from tqdm import tqdm

cartella = "/srv/data/arcangelo/meta_output_current/csv_out"

def get_csv_data(filepath: str) -> list:
    if not os.path.splitext(filepath)[1].endswith('.csv'):
        return []

    cur_field_size = 128
    field_size_changed = False
    data = []

    while not data:
        try:
            with open(filepath, 'r', encoding='utf8') as data_initial:
                valid_data = (line.replace('\0', '').replace('\n', '') for line in data_initial)
                data = list(csv.DictReader(valid_data, delimiter=','))
        except csv.Error:
            cur_field_size *= 2
            csv.field_size_limit(cur_field_size)
            field_size_changed = True

    if field_size_changed:
        csv.field_size_limit(128)
    return data

def get_full_language_name(lang_code):
    try:
        return pycountry.languages.get(alpha_2=lang_code).name
    except AttributeError:
        return lang_code  # Se il codice non viene trovato, restituisci il codice originale

def process_file(file):
    data = get_csv_data(os.path.join(cartella, file))
    local_lingue = []

    for riga in data:
        try:
            title = riga['title']
            lingua = detect(title)
            full_lang_name = get_full_language_name(lingua)
            local_lingue.append(full_lang_name)
        except:
            pass

    return local_lingue

def main():
    # Crea una lista dei file csv nella cartella
    file_csv = [f for f in os.listdir(cartella) if f.endswith('.csv')]

    results = []

    totale_file = len(file_csv)

    with ProcessPoolExecutor() as executor:
        results = list(tqdm(executor.map(process_file, file_csv), total=totale_file, desc="Elaborazione files"))

    lingue = [lang for sublist in results for lang in sublist]

    # Conta quante volte ogni lingua è stata rilevata
    conteggio = Counter(lingue)

    # Salva i risultati in un file CSV
    with open('/srv/data/arcangelo/meta_output_current/lingue_in_meta.csv', 'w', newline='', encoding='utf-8') as f:
        writer = csv.writer(f)
        writer.writerow(['Lingua', 'Conteggio'])
        for lingua, count in conteggio.items():
            writer.writerow([lingua, count])
    print("Elaborazione completata e risultati salvati in risultati.csv!")\

if __name__ == '__main__':
    main()


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from typing import Tuple

from oc_meta.plugins.csv_generator.csv_generator import (find_file,
                                                         process_archive,
                                                         process_id)


def extract_identifiers(data: list, filepath: str, rdf_dir:str, dir_split_number: int, items_per_file: int, zip_output_rdf: bool, memory: dict, meta_config: str, resp_agent: str) -> Tuple[str, list]:
    to_be_merged = list()
    for graph in data:
        graph_data = graph['@graph']
        for entity in graph_data:
            if 'http://purl.org/spar/datacite/hasIdentifier' in entity:
                prev_id_uris = dict()
                identifiers = sorted([identifier['@id'] for identifier in entity['http://purl.org/spar/datacite/hasIdentifier']])
                for id_uri in identifiers:
                    id_path = find_file(rdf_dir, dir_split_number, items_per_file, id_uri, zip_output_rdf)
                    id_full = process_archive(id_path, process_id, memory, id_uri, meta_config, resp_agent, id_path, memory, entity)
                    if id_full in prev_id_uris:
                        prev_id_uri = prev_id_uris[id_full]
                        to_be_merged.append(tuple([prev_id_uri, id_uri]))
                    else:
                        prev_id_uris[id_full] = id_uri
    return filepath, to_be_merged


from oc_meta.lib.file_manager import get_csv_data, write_csv


def generate_table_from_list_of_ids(filepath:str):
    table = []
    ids = set()
    data = get_csv_data(filepath)
    for row in data:
        ids.add(row['id'])
    for identifier in ids:
        fieldnames = ['id', 'title', 'author', 'issue', 'volume', 'venue', 'pub_date', 'page', 'type', 'editor', 'publisher']
        fields_dict = {field: '' for field in fieldnames}
        fields_dict['id'] = f"doi:{identifier}"
        table.append(fields_dict)
    write_csv('E:/ivan_missing/1.csv', table)

generate_table_from_list_of_ids('E:/ivan_missing/coci_entities_with_no_omid_06_2022.csv')


#!python
# Copyright 2022, Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from __future__ import annotations

import os
import re
from datetime import datetime
from functools import cmp_to_key
from typing import Dict, List

from dateutil.parser import parse
from tqdm import tqdm

from oc_meta.lib.file_manager import get_csv_data, write_csv
from oc_meta.lib.master_of_regex import name_and_ids


class OCMetaAnalyser:
    def __init__(self, csv_dump_path: str):
        self.csv_dump_path = csv_dump_path
    
    def merge_rows_by_id(self, output_dir: str) -> None:
        ids_by_csv = dict()
        for filename in os.listdir(self.csv_dump_path):
            csv_data = get_csv_data(os.path.join(self.csv_dump_path, filename))
            for i, row in enumerate(csv_data):
                metaid = [identifier for identifier in row['id'].split() if identifier.split(':', maxsplit=1)[0] == 'omid'][0]
                ids_by_csv.setdefault(metaid, dict())
                ids_by_csv[metaid].setdefault(filename, set())
                ids_by_csv[metaid][filename].add(i)
        storer = dict()
        for metaid, filenames in ids_by_csv.items():
            if len(filenames) > 1:
                sorted_filenames = sorted([name for name in filenames], key=cmp_to_key(self.sort_csv_filenames))
                to_be_overritten = sorted_filenames[:-1]
                latest_file = sorted_filenames[-1]
                for filename in to_be_overritten:
                    storer.setdefault(filename, set())
                    storer[filename].update(ids_by_csv[metaid][filename])
                storer.setdefault(latest_file, set())
            else:
                storer.setdefault(list(filenames.keys())[0], set())
        for filename, rows in storer.items():
            old_data = get_csv_data(os.path.join(self.csv_dump_path, filename))
            new_data = [row for i, row in enumerate(old_data) if i not in rows]
            write_csv(
                path=os.path.join(output_dir, filename), 
                datalist=new_data, 
                fieldnames=['id', 'title', 'pub_date', 'page', 'type', 'author', 'editor', 'publisher', 'volume', 'venue', 'issue'], 
                method='w')
            
    @staticmethod
    def sort_csv_filenames(file_1, file_2) -> str:
        file_1_date = datetime.strptime(file_1.split('_')[1].replace('.csv', ''), '%Y-%m-%dT%H-%M-%S')
        file_2_date = datetime.strptime(file_2.split('_')[1].replace('.csv', ''), '%Y-%m-%dT%H-%M-%S')
        if file_1_date > file_2_date:
            return 1
        elif file_1_date < file_2_date:
            return -1
        elif file_1_date == file_2_date:
            if int(file_1.split('_')[0]) > int(file_2.split('_')[0]):
                return 1
            else:
                return -1

    def explore_csv_dump(self, analyser: callable) -> None|int|dict:
        global_output = None
        filenames = sorted(os.listdir(self.csv_dump_path))
        pbar = tqdm(total=len(filenames))
        for i, filename in enumerate(filenames):
            csv_data = get_csv_data(os.path.join(self.csv_dump_path, filename))
            local_output = analyser(csv_data)
            if i == 0:
                if isinstance(local_output, int):
                    global_output = 0
                elif isinstance(local_output, dict):
                    global_output = dict()
                elif isinstance(local_output, set):
                    global_output = set()
            if isinstance(local_output, int):
                global_output += local_output
            elif isinstance(local_output, dict):
                for k,v in local_output.items():
                    if k in global_output:
                        for i_k, _ in v.items():
                            if i_k in global_output[k]:
                                if isinstance(global_output[k][i_k], set):
                                    global_output[k][i_k].update(local_output[k][i_k])
                            else:
                                global_output[k][i_k] = local_output[k][i_k]
                    else:
                        global_output[k] = local_output[k]
            elif isinstance(local_output, set):
                global_output.update(local_output)      
            pbar.update()
        pbar.close()
        if isinstance(global_output, int):
            return str(global_output)
        elif isinstance(global_output, dict):
            return global_output
        elif isinstance(global_output, set):
            return str(len(global_output))
    

class OCMetaCounter(OCMetaAnalyser):
    def __init__(self, csv_dump_path: str):
        super(OCMetaCounter, self).__init__(csv_dump_path)
    
    def get_top(self, what: str, by_what: str, number: int|None = None) -> dict:
        counter_func = getattr(self, f'count_{what}_by_{by_what}')
        all_data = self.explore_csv_dump(counter_func)
        all_data_sorted: list = sorted(all_data, key=lambda k: len(all_data[k][by_what]), reverse=True)
        top_n = all_data_sorted[:number] if number is not None else all_data_sorted
        all_top_n = [(k, v) for k, v in all_data.items() if k in top_n]
        for tuple_k_v in all_top_n:
            tuple_k_v[1]['total'] = len(tuple_k_v[1][by_what])
        all_top_n = [(meta, {k: v for k, v in data.items() if not isinstance(v, set)}) for meta, data in all_top_n]
        return sorted(all_top_n, key=lambda x: top_n.index(x[0]))
    
    def count(self, what: str) -> int:
        counter_func = getattr(self, f'count_{what}')
        return self.explore_csv_dump(counter_func)

    def count_authors(self, csv_data: List[dict]) -> int:
        count = 0
        for row in csv_data:
            count += len(list(filter(None, row['author'].split('; '))))
        return count

    def count_editors(self, csv_data: List[dict]) -> int:
        count = 0
        for row in csv_data:
            count += len(list(filter(None, row['editor'].split('; '))))
        return count

    def count_publishers(self, csv_data: List[dict]) -> set:
        publishers = set()
        for row in csv_data:
            if row['publisher']:
                pub_name_and_ids = re.search(name_and_ids, row['publisher'])
                if pub_name_and_ids:
                    pub_name = pub_name_and_ids.group(1).lower()
                    publishers.add(pub_name)
        return publishers

    def count_venues(self, csv_data: List[dict]) -> set:
        venues = set()
        for row in csv_data:
            if row['venue']:
                ven_name_and_ids = re.search(name_and_ids, row['venue'])
                venue_name = ven_name_and_ids.group(1).lower()
                venue_ids = set(ven_name_and_ids.group(2).split())
                venue_metaid = [identifier for identifier in venue_ids if identifier.split(':', maxsplit=1)[0] == 'omid'][0]
                if not venue_ids.difference({venue_metaid}):
                    venues.add(venue_name)
                else:
                    venues.add(venue_metaid)
        return venues
    
    def count_publishers_by_venue(self, csv_data: List[dict]) -> Dict[str, Dict[str, set|str]]:
        publishers_by_venue = dict()
        for row in csv_data:
            publisher_name_and_ids = re.search(name_and_ids, row['publisher'])
            venue_name_and_ids = re.search(name_and_ids, row['venue'])
            if publisher_name_and_ids and venue_name_and_ids:
                publisher_name = publisher_name_and_ids.group(1).lower()
                venue_name: str = venue_name_and_ids.group(1).lower()
                venue_ids = set(venue_name_and_ids.group(2).split())
                venue_metaid = [identifier for identifier in venue_ids if identifier.split(':', maxsplit=1)[0] == 'omid'][0]
                publishers_by_venue.setdefault(publisher_name, {'name': publisher_name, 'venue': set()})
                venue_key = venue_name if not venue_ids.difference({venue_metaid}) else venue_metaid
                publishers_by_venue[publisher_name]['venue'].add(venue_key)
                return publishers_by_venue

    def count_publishers_by_publication(self, csv_data: List[dict]) -> Dict[str, Dict[str, set|str]]:
        publishers_by_publication = dict()
        for row in csv_data:
            publishers_name_and_ids = re.search(name_and_ids, row['publisher'])
            if publishers_name_and_ids:
                publishers_name = publishers_name_and_ids.group(1)
                row_metaid = [identifier for identifier in row['id'].split() if identifier.split(':', maxsplit=1)[0] == 'omid'][0]
                publishers_by_publication.setdefault(publishers_name.lower(), {'name': publishers_name, 'publication': set()})
                publishers_by_publication[publishers_name.lower()]['publication'].add(row_metaid)
        return publishers_by_publication

    def count_venues_by_publication(self, csv_data: List[dict]) -> Dict[str, Dict[str, set|str]]:
        venues_by_publication = dict()
        for row in csv_data:
            venue_name_and_ids = re.search(name_and_ids, row['venue'])
            if venue_name_and_ids:
                venue_name = venue_name_and_ids.group(1)
                venue_ids = set(venue_name_and_ids.group(2).split())
                venue_metaid = [identifier for identifier in venue_ids if identifier.split(':', maxsplit=1)[0] == 'omid'][0]
                row_metaid = [identifier for identifier in row['id'].split() if identifier.split(':', maxsplit=1)[0] == 'omid'][0]
                venue_key = venue_name.lower() if not venue_ids.difference({venue_metaid}) else venue_metaid
                venues_by_publication.setdefault(venue_key, {'name': venue_name, 'publication': set()})
                venues_by_publication[venue_key]['publication'].add(row_metaid)
        return venues_by_publication

    def count_years_by_publication(self, csv_data: List[dict]) -> Dict[str, Dict[str, set]]:
        years_by_publication = dict()
        for row in csv_data:
            pub_date = row['pub_date']
            if pub_date:
                year = datetime.strftime(parse(pub_date), '%Y')
                row_metaid = [identifier for identifier in row['id'].split() if identifier.split(':', maxsplit=1)[0] == 'omid'][0]
                years_by_publication.setdefault(year, {'publication': set()})
                years_by_publication[year]['publication'].add(row_metaid)
        return years_by_publication

    def count_types_by_publication(self, csv_data: List[dict]) -> Dict[str, Dict[str, set|str]]:
        types_by_publication = dict()
        for row in csv_data:
            br_type = row['type']
            if br_type:
                row_metaid = [identifier for identifier in row['id'].split() if identifier.split(':', maxsplit=1)[0] == 'omid'][0]
                types_by_publication.setdefault(br_type, {'publication': set()})
                types_by_publication[br_type]['publication'].add(row_metaid)
                venue_name_and_ids = re.search(name_and_ids, row['venue'])
                if venue_name_and_ids:
                    venue_name = venue_name_and_ids.group(1)
                    venue_ids = set(venue_name_and_ids.group(2).split())
                    venue_type = self.get_venue_type(br_type, venue_ids)
                    venue_metaid = [identifier for identifier in venue_ids if identifier.split(':', maxsplit=1)[0] == 'omid'][0]
                    if venue_type:
                        if not venue_ids.difference({venue_metaid}):
                            venue_key = venue_name
                        else:
                            venue_key = venue_metaid
                        types_by_publication.setdefault(venue_type, {'publication': set()})
                        types_by_publication[venue_type]['publication'].add(venue_key)
        return types_by_publication

    @classmethod
    def get_venue_type(cls, br_type:str, venue_ids:list) -> str:
        schemas = {venue_id.split(':', maxsplit=1)[0] for venue_id in venue_ids}
        if br_type in {'journal article', 'journal volume', 'journal issue'}:
            venue_type = 'journal'
        elif br_type in {'book chapter', 'book part', 'book section', 'book track'}:
            venue_type = 'book'
        elif br_type in {'book', 'edited book', 'monograph', 'reference book'}:
            venue_type = 'book series'
        elif br_type == 'proceedings article':
            venue_type = 'proceedings'
        elif br_type in {'proceedings', 'report', 'standard', 'series'}:
            venue_type = 'series'
        elif br_type == 'reference entry':
            venue_type = 'reference book'
        elif br_type == 'report series':
            venue_type = 'report series'
        elif not br_type or br_type in {'dataset', 'data file', 'journal'}:
            venue_type = ''
        # Check the type based on the identifier scheme
        if any(identifier for identifier in venue_ids if not identifier.startswith('omid:')):
            try:
                if venue_type in {'journal', 'book series', 'series', 'report series'}:
                    if 'isbn' in schemas or 'issn' not in schemas:
                        # It is undecidable
                        venue_type = ''
                elif venue_type in {'book', 'proceedings'}:
                    if 'issn' in schemas or 'isbn' not in schemas:
                        venue_type = ''
                elif venue_type == 'reference book':
                    if 'isbn' in schemas and 'issn' not in schemas:
                        venue_type = 'reference book'
                    elif 'issn' in schemas and 'isbn' not in schemas:
                        venue_type = 'journal'
                    elif 'issn' in schemas and 'isbn' in schemas:
                        venue_type = ''
            except UnboundLocalError:
                print(br_type, venue_ids)
                raise(UnboundLocalError)
        return venue_type


import argparse
import zipfile
import os
import json
from rdflib import ConjunctiveGraph, URIRef
from tqdm import tqdm
from multiprocessing import Pool, cpu_count

def get_entity_type(entity_url):
    parts = entity_url.split('/')
    if 'oc' in parts and 'meta' in parts:
        try:
            return parts[parts.index('meta') + 1]
        except IndexError:
            return None
    return None

def process_json_content(json_content):
    entity_count = {}
    total_entities = {}
    for item in json_content:
        graph = item["@graph"]
        for entity in graph:
            entity_id = entity["@id"]
            entity_type = get_entity_type(entity_id)
            if entity_type:
                if entity_type in total_entities:
                    total_entities[entity_type] += 1
                else:
                    total_entities[entity_type] = 1

                if "http://www.w3.org/ns/prov#hadPrimarySource" not in entity:
                    if entity_type in entity_count:
                        entity_count[entity_type] += 1
                    else:
                        entity_count[entity_type] = 1

    return entity_count, total_entities

def process_zip_file(zip_path):
    with zipfile.ZipFile(zip_path, 'r') as z:
        file_info = z.infolist()[0]
        if file_info.filename.endswith('.json'):
            with z.open(file_info.filename) as file:
                json_content = json.load(file)
                entity_count, total_count = process_json_content(json_content)
                return entity_count, total_count
    return {}, {}

def combine_counts(main_count, new_count):
    for key, value in new_count.items():
        if key in main_count:
            main_count[key] += value
        else:
            main_count[key] = value
    return main_count

def find_and_process_zip_files(folder_path):
    zip_files = [os.path.join(root, file)
                 for root, _, files in os.walk(folder_path)
                 for file in files if file == 'se.zip']
    pool = Pool(cpu_count())
    results = list(tqdm(pool.imap(process_zip_file, zip_files), total=len(zip_files)))
    pool.close()
    pool.join()

    final_count = {}
    total_count = {}
    for entity_count, total_entities in results:
        final_count = combine_counts(final_count, entity_count)
        total_count = combine_counts(total_count, total_entities)

    with open("missing_primary_source_log.txt", "w") as log_file:
        for entity_type, missing_count in final_count.items():
            total = total_count.get(entity_type, 0)
            percentage = (missing_count / total * 100) if total > 0 else 0
            log_entry = f"Entity Type: {entity_type}, Missing 'hadPrimarySource': {missing_count}/{total} ({percentage:.2f}%)"
            print(log_entry)
            log_file.write(log_entry + "\n")

def main():
    parser = argparse.ArgumentParser(description="Explore and process se.zip files for specific RDF data.")
    parser.add_argument("folder_path", type=str, help="Root folder to search for se.zip files.")
    args = parser.parse_args()
    find_and_process_zip_files(args.folder_path)

if __name__ == "__main__":
    main()



from rdflib import URIRef
from SPARQLWrapper import JSON, SPARQLWrapper

from oc_meta.plugins.editor import MetaEditor

server = SPARQLWrapper('http://127.0.0.1:19999/blazegraph/sparql')
query = '''
PREFIX pro: <http://purl.org/spar/pro/>
SELECT DISTINCT ?br ?ar
WHERE {
    <https://w3id.org/oc/meta/ra/0630115031> ^pro:isHeldBy ?ar.
    ?br pro:isDocumentContextFor ?ar.
}
'''
server.setQuery(query)
server.setReturnFormat(JSON)
result = server.queryAndConvert()
br_ar = dict()
meta_editor = MetaEditor('E:\meta_output_24_04_2023_datacite_fixes\meta_output\meta_config.yaml', 'https://orcid.org/0000-0002-8420-0696')
for r in result['results']['bindings']:
    br = r['br']['value']
    br_ar.setdefault(br, list())
    ar = r['ar']['value']
    br_ar[br].append(ar)
for br, ars in br_ar.items():
    if len(ars) == 2:
        meta_editor.merge(URIRef(ars[0]), URIRef(ars[1]))


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2022, Silvio Peroni <essepuntato@gmail.com>, Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

from __future__ import annotations

import gzip
import os.path
import pathlib
import tarfile
import zipfile
from json import load, loads
from os import makedirs, sep, walk
from os.path import basename, exists, isdir
from typing import Tuple

import zstandard as zstd

from oc_meta.lib.file_manager import init_cache


def get_all_files(is_dir_or_targz_file:str, cache_filepath:str|None=None) -> Tuple[list, tarfile.TarFile|None]:
    result = []
    targz_fd = None
    cache = init_cache(cache_filepath)
    if isdir(is_dir_or_targz_file):
        for cur_dir, _, cur_files in walk(is_dir_or_targz_file):
            for cur_file in cur_files:
                if (cur_file.endswith(".json") or cur_file.endswith(".json.gz")) and not basename(cur_file).startswith(".") and not cur_file in cache:
                    result.append(cur_dir + sep + cur_file)
    elif is_dir_or_targz_file.endswith("tar.gz"):
        targz_fd = tarfile.open(is_dir_or_targz_file, "r:gz", encoding="utf-8")
        for cur_file in targz_fd:
            if cur_file.name.endswith(".json") and not basename(cur_file.name).startswith(".") and not cur_file.name in cache:
                result.append(cur_file)
    else:
        print("It is not possible to process the input path.")
    return result, targz_fd

def load_json(file:str|tarfile.TarInfo, targz_fd:tarfile.TarFile) -> dict|None:
    result = None
    if targz_fd is None:
        if file.endswith(".json"):  # type: ignore
            with open(file, encoding="utf8") as f: # type: ignore
                result = load(f)
        elif file.endswith(".json.gz"): # type: ignore
            with gzip.open(file, 'r') as gzip_file: # type: ignore
                data = gzip_file.read()
                result = loads(data.decode('utf-8'))
    else:
        cur_tar_file = targz_fd.extractfile(file)
        json_str = cur_tar_file.read() # type: ignore
        # In Python 3.5 it seems that, for some reason, the extractfile method returns an
        # object 'bytes' that cannot be managed by the function 'load' in the json package.
        # Thus, to avoid issues, in case an object having type 'bytes' is return, it is
        # transformed as a string before passing it to the function 'loads'. Please note
        # that Python 3.9 does not show this behaviour, and it works correctly without
        # any transformation.
        if type(json_str) is bytes:
            json_str = json_str.decode("utf-8")
        result = loads(json_str)
    return result


def get_all_files_by_type(i_dir_or_compr:str, req_type:str, cache_filepath:str|None=None):
    result = []
    targz_fd = None
    cache = init_cache(cache_filepath)

    if isdir(i_dir_or_compr):

        for cur_dir, cur_subdir, cur_files in walk(i_dir_or_compr):
            for cur_file in cur_files:
                if cur_file.endswith(req_type) and not basename(cur_file).startswith(".") and not cur_file in cache:
                    result.append(os.path.join(cur_dir, cur_file))
    elif i_dir_or_compr.endswith("tar.gz"):
        targz_fd = tarfile.open(i_dir_or_compr, "r:gz", encoding="utf-8")
        for cur_file in targz_fd:
            if cur_file.name.endswith(req_type) and not basename(cur_file.name).startswith(".") and not cur_file in cache:
                result.append(cur_file)
    elif i_dir_or_compr.endswith("zip"):
        with zipfile.ZipFile(i_dir_or_compr, 'r') as zip_ref:
            dest_dir = i_dir_or_compr.replace('.zip', '') + "decompr_zip_dir"
            if not exists(dest_dir):
                makedirs(dest_dir)
            zip_ref.extractall(dest_dir)
        for cur_dir, cur_subdir, cur_files in walk(dest_dir):
            for cur_file in cur_files:
                if cur_file.endswith(req_type) and not basename(cur_file).startswith(".") and not cur_file in cache:
                    result.append(cur_dir + sep + cur_file)

    elif i_dir_or_compr.endswith("zst"):
        input_file = pathlib.Path(i_dir_or_compr)
        dest_dir = i_dir_or_compr.split(".")[0] + "_decompr_zst_dir"
        with open(input_file, 'rb') as compressed:
            decomp = zstd.ZstdDecompressor()
            if not exists(dest_dir):
                makedirs(dest_dir)
            output_path = pathlib.Path(dest_dir) / input_file.stem
            if not exists(output_path):
                with open(output_path, 'wb') as destination:
                    decomp.copy_stream(compressed, destination)
        for cur_dir, cur_subdir, cur_files in walk(dest_dir):
            for cur_file in cur_files:
                if cur_file.endswith(req_type) and not basename(cur_file).startswith(".") and not cur_file in cache:
                    result.append(cur_dir + sep + cur_file)
    else:
        print("It is not possible to process the input path.", i_dir_or_compr)
    return result, targz_fd
